# Copyright 2016 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ctypes

import ddt
import mock
from six.moves import queue

from os_win import constants
from os_win import exceptions
from os_win.tests.unit import test_base
from os_win.utils.compute import clusterutils
from os_win.utils.winapi import constants as w_const
from os_win.utils.winapi.libs import clusapi as clusapi_def
from os_win.utils.winapi import wintypes


@ddt.ddt
class ClusterUtilsTestCase(test_base.OsWinBaseTestCase):
    """Unit tests for the Hyper-V ClusterUtilsBase class."""

    _autospec_classes = [
        clusterutils._clusapi_utils.ClusApiUtils,
        clusterutils._clusapi_utils.ClusterContextManager
    ]

    _FAKE_RES_NAME = "fake_res_name"
    _FAKE_HOST = "fake_host"
    _FAKE_PREV_HOST = "fake_prev_host"
    _FAKE_VM_NAME = 'instance-00000001'
    _FAKE_RESOURCEGROUP_NAME = 'Virtual Machine %s' % _FAKE_VM_NAME

    def setUp(self):
        super(ClusterUtilsTestCase, self).setUp()
        self._clusterutils = clusterutils.ClusterUtils()
        self._clusterutils._conn_cluster = mock.MagicMock()
        self._clusterutils._cluster = mock.MagicMock()
        self._clusapi = self._clusterutils._clusapi_utils
        self._cmgr = self._clusterutils._cmgr

    def _cmgr_val(self, cmgr):
        # Return the value that a mocked context manager would yield.
        return cmgr.return_value.__enter__.return_value

    def test_init_hyperv_conn(self):
        fake_cluster_name = "fake_cluster"
        mock_cluster = mock.MagicMock()
        mock_cluster.path_.return_value = r"\\%s\root" % fake_cluster_name

        mock_conn = mock.MagicMock()
        mock_conn.MSCluster_Cluster.return_value = [mock_cluster]

        self._clusterutils._get_wmi_conn = mock.MagicMock()
        self._clusterutils._get_wmi_conn.return_value = mock_conn

        self._clusterutils._init_hyperv_conn("fake_host")

    def test_init_hyperv_conn_exception(self):
        self._clusterutils._get_wmi_conn = mock.MagicMock()
        self._clusterutils._get_wmi_conn.side_effect = AttributeError
        self.assertRaises(exceptions.HyperVClusterException,
                          self._clusterutils._init_hyperv_conn, "fake_host")

    @mock.patch.object(clusterutils.ClusterUtils,
                       '_get_cluster_nodes')
    def test_check_cluster_state_not_enough_nodes(self, mock_get_nodes):
        self.assertRaises(exceptions.HyperVClusterException,
                          self._clusterutils.check_cluster_state)

    def test_get_node_name(self):
        self._clusterutils._this_node = mock.sentinel.fake_node_name
        self.assertEqual(mock.sentinel.fake_node_name,
                         self._clusterutils.get_node_name())

    @mock.patch.object(clusterutils.ClusterUtils, 'cluster_enum')
    def test_get_cluster_nodes(self, mock_cluster_enum):
        expected = mock_cluster_enum.return_value

        self.assertEqual(expected, self._clusterutils._get_cluster_nodes())

        mock_cluster_enum.assert_called_once_with(w_const.CLUSTER_ENUM_NODE)

    @mock.patch.object(clusterutils.ClusterUtils, 'cluster_enum')
    @mock.patch.object(clusterutils.ClusterUtils, 'get_cluster_group_type')
    def test_get_vm_groups(self, mock_get_type, mock_cluster_enum):
        mock_groups = [mock.MagicMock(), mock.MagicMock(), mock.MagicMock()]
        group_types = [w_const.ClusGroupTypeVirtualMachine,
                       w_const.ClusGroupTypeVirtualMachine,
                       mock.sentinel.some_other_group_type]

        mock_cluster_enum.return_value = mock_groups
        mock_get_type.side_effect = group_types

        exp = mock_groups[:-1]
        res = list(self._clusterutils._get_vm_groups())

        self.assertEqual(exp, res)

        mock_cluster_enum.assert_called_once_with(w_const.CLUSTER_ENUM_GROUP)
        mock_get_type.assert_has_calls(
            [mock.call(r['name']) for r in mock_groups])

    @mock.patch.object(clusterutils.ClusterUtils,
                       '_lookup_vm_group')
    def test_lookup_vm_group_check(self, mock_lookup_vm_group):
        mock_lookup_vm_group.return_value = mock.sentinel.fake_vm

        ret = self._clusterutils._lookup_vm_group_check(
            self._FAKE_VM_NAME)
        self.assertEqual(mock.sentinel.fake_vm, ret)

    @mock.patch.object(clusterutils.ClusterUtils,
                       '_lookup_vm_group')
    def test_lookup_vm_group_check_no_vm(self, mock_lookup_vm_group):
        mock_lookup_vm_group.return_value = None

        self.assertRaises(exceptions.HyperVVMNotFoundException,
                          self._clusterutils._lookup_vm_group_check,
                          self._FAKE_VM_NAME)

    @mock.patch.object(clusterutils.ClusterUtils,
                       '_lookup_res')
    def test_lookup_vm_group(self, mock_lookup_res):
        self._clusterutils._lookup_vm_group(self._FAKE_VM_NAME)
        mock_lookup_res.assert_called_once_with(
            self._clusterutils._conn_cluster.MSCluster_ResourceGroup,
            self._FAKE_VM_NAME)

    def test_lookup_res_no_res(self):
        res_list = []
        resource_source = mock.MagicMock()
        resource_source.return_value = res_list

        self.assertIsNone(
            self._clusterutils._lookup_res(resource_source,
                                           self._FAKE_RES_NAME))
        resource_source.assert_called_once_with(
            Name=self._FAKE_RES_NAME)

    def test_lookup_res_duplicate_res(self):
        res_list = [mock.sentinel.r1,
                    mock.sentinel.r1]
        resource_source = mock.MagicMock()
        resource_source.return_value = res_list

        self.assertRaises(exceptions.HyperVClusterException,
                          self._clusterutils._lookup_res,
                          resource_source,
                          self._FAKE_RES_NAME)
        resource_source.assert_called_once_with(
            Name=self._FAKE_RES_NAME)

    def test_lookup_res(self):
        res_list = [mock.sentinel.r1]
        resource_source = mock.MagicMock()
        resource_source.return_value = res_list

        self.assertEqual(
            mock.sentinel.r1,
            self._clusterutils._lookup_res(resource_source,
                                           self._FAKE_RES_NAME))
        resource_source.assert_called_once_with(
            Name=self._FAKE_RES_NAME)

    @mock.patch.object(clusterutils.ClusterUtils,
                       '_get_cluster_nodes')
    def test_get_cluster_node_names(self, mock_get_cluster_nodes):
        cluster_nodes = [dict(name='node1'),
                         dict(name='node2')]
        mock_get_cluster_nodes.return_value = cluster_nodes

        ret = self._clusterutils.get_cluster_node_names()

        self.assertItemsEqual(['node1', 'node2'], ret)

    @mock.patch.object(clusterutils.ClusterUtils, '_get_cluster_group_state')
    def test_get_vm_host(self, mock_get_state):
        # Refresh the helpers. Closures are a bit difficult to mock.
        owner_node = "fake_owner_node"
        mock_get_state.return_value = dict(owner_node=owner_node)

        self.assertEqual(
            owner_node,
            self._clusterutils.get_vm_host(mock.sentinel.vm_name))

        self._cmgr.open_cluster_group.assert_called_once_with(
            mock.sentinel.vm_name)
        mock_get_state.assert_called_once_with(
            self._cmgr_val(self._cmgr.open_cluster_group))

    @mock.patch.object(clusterutils.ClusterUtils, '_get_vm_groups')
    def test_list_instances(self, mock_get_vm_groups):
        mock_get_vm_groups.return_value = [dict(name='vm1'),
                                           dict(name='vm2')]
        ret = self._clusterutils.list_instances()
        self.assertItemsEqual(['vm1', 'vm2'], ret)

    @mock.patch.object(clusterutils.ClusterUtils, '_get_vm_groups')
    def test_list_instance_uuids(self, mock_get_vm_groups):
        mock_get_vm_groups.return_value = [dict(id='uuid1'),
                                           dict(id='uuid2')]
        ret = self._clusterutils.list_instance_uuids()
        self.assertItemsEqual(['uuid1', 'uuid2'], ret)

    @ddt.data(True, False)
    @mock.patch.object(clusterutils.ClusterUtils,
                       '_lookup_vm_group_check')
    def test_add_vm_to_cluster(self, auto_failback,
                               mock_lookup_vm_group_check):
        self._clusterutils._cluster.AddVirtualMachine = mock.MagicMock()
        vm_group = mock.Mock()
        mock_lookup_vm_group_check.return_value = vm_group

        self._clusterutils.add_vm_to_cluster(
            self._FAKE_VM_NAME, mock.sentinel.max_failover_count,
            mock.sentinel.failover_period, auto_failback)

        self.assertEqual(mock.sentinel.max_failover_count,
                         vm_group.FailoverThreshold)
        self.assertEqual(mock.sentinel.failover_period,
                         vm_group.FailoverPeriod)
        self.assertTrue(vm_group.PersistentState)
        self.assertEqual(vm_group.AutoFailbackType, int(auto_failback))
        self.assertEqual(vm_group.FailbackWindowStart,
                         self._clusterutils._FAILBACK_WINDOW_MIN)
        self.assertEqual(vm_group.FailbackWindowEnd,
                         self._clusterutils._FAILBACK_WINDOW_MAX)
        vm_group.put.assert_called_once_with()

    def test_bring_online(self):
        self._clusterutils.bring_online(mock.sentinel.vm_name)

        self._cmgr.open_cluster_group.assert_called_once_with(
            mock.sentinel.vm_name)
        self._clusapi.online_cluster_group.assert_called_once_with(
            self._cmgr_val(self._cmgr.open_cluster_group))

    def test_take_offline(self):
        self._clusterutils.take_offline(mock.sentinel.vm_name)

        self._cmgr.open_cluster_group.assert_called_once_with(
            mock.sentinel.vm_name)
        self._clusapi.offline_cluster_group.assert_called_once_with(
            self._cmgr_val(self._cmgr.open_cluster_group))

    @mock.patch.object(clusterutils.ClusterUtils, '_lookup_vm_group')
    def test_delete(self, mock_lookup_vm_group):
        vm = mock.MagicMock()
        mock_lookup_vm_group.return_value = vm

        self._clusterutils.delete(self._FAKE_VM_NAME)
        vm.DestroyGroup.assert_called_once_with(
            self._clusterutils._DESTROY_GROUP)

    def test_cluster_enum(self):
        cluster_objects = [mock.Mock(), mock.Mock()]

        self._clusapi.cluster_get_enum_count.return_value = len(
            cluster_objects)
        self._clusapi.cluster_enum.side_effect = cluster_objects

        exp_ret_val = [dict(version=item.dwVersion,
                            type=item.dwType,
                            id=item.lpszId,
                            name=item.lpszName) for item in cluster_objects]
        ret_val = list(self._clusterutils.cluster_enum(mock.sentinel.obj_type))

        self.assertEqual(exp_ret_val, ret_val)

        enum_handle = self._cmgr_val(self._cmgr.open_cluster_enum)
        self._cmgr.open_cluster_enum.assert_called_once_with(
            mock.sentinel.obj_type)
        self._clusapi.cluster_get_enum_count.assert_called_once_with(
            enum_handle)
        self._clusapi.cluster_enum.assert_has_calls(
            [mock.call(enum_handle, idx)
             for idx in range(len(cluster_objects))])

    @ddt.data(True, False)
    def test_vm_exists(self, exists):
        self._cmgr.open_cluster_resource.side_effect = (
            None if exists else exceptions.ClusterObjectNotFound('test'))

        self.assertEqual(
            exists,
            self._clusterutils.vm_exists(self._FAKE_VM_NAME))

        self._cmgr.open_cluster_resource.assert_called_once_with(
            self._FAKE_RESOURCEGROUP_NAME)

    @mock.patch.object(clusterutils.ClusterUtils, '_migrate_vm')
    def test_live_migrate_vm(self, mock_migrate_vm):
        self._clusterutils.live_migrate_vm(self._FAKE_VM_NAME,
                                           self._FAKE_HOST,
                                           mock.sentinel.timeout)

        mock_migrate_vm.assert_called_once_with(
            self._FAKE_VM_NAME, self._FAKE_HOST,
            self._clusterutils._LIVE_MIGRATION_TYPE,
            constants.CLUSTER_GROUP_ONLINE,
            mock.sentinel.timeout)

    @mock.patch.object(wintypes, 'DWORD')
    @mock.patch.object(clusterutils.ClusterUtils,
                       '_wait_for_cluster_group_migration')
    @mock.patch.object(clusterutils.ClusterUtils,
                       '_validate_migration')
    @mock.patch.object(clusterutils,
                       '_ClusterGroupStateChangeListener')
    @ddt.data(None, exceptions.ClusterException)
    def test_migrate_vm(self, wait_unexpected_exc,
                        mock_listener_cls,
                        mock_validate_migr,
                        mock_wait_group, mock_dword):
        mock_wait_group.side_effect = wait_unexpected_exc

        migrate_args = (self._FAKE_VM_NAME,
                        self._FAKE_HOST,
                        self._clusterutils._LIVE_MIGRATION_TYPE,
                        constants.CLUSTER_GROUP_ONLINE,
                        mock.sentinel.timeout)

        if wait_unexpected_exc:
            self.assertRaises(wait_unexpected_exc,
                              self._clusterutils._migrate_vm,
                              *migrate_args)
        else:
            self._clusterutils._migrate_vm(*migrate_args)

        mock_dword.assert_called_once_with(
            self._clusterutils._LIVE_MIGRATION_TYPE)

        self._clusapi.get_property_list_entry.assert_has_calls(
            [mock.call(prop_name,
                       w_const.CLUSPROP_SYNTAX_LIST_VALUE_DWORD,
                       mock_dword.return_value)
             for prop_name in (w_const.CLUS_RESTYPE_NAME_VM,
                               w_const.CLUS_RESTYPE_NAME_VM_CONFIG)])

        expected_prop_entries = [
            self._clusapi.get_property_list_entry.return_value] * 2
        self._clusapi.get_property_list.assert_called_once_with(
            expected_prop_entries)

        expected_migrate_flags = (
            w_const.CLUSAPI_GROUP_MOVE_RETURN_TO_SOURCE_NODE_ON_ERROR |
            w_const.CLUSAPI_GROUP_MOVE_QUEUE_ENABLED |
            w_const.CLUSAPI_GROUP_MOVE_HIGH_PRIORITY_START)

        exp_clus_h = self._cmgr_val(self._cmgr.open_cluster)
        exp_clus_node_h = self._cmgr_val(self._cmgr.open_cluster_node)
        exp_clus_group_h = self._cmgr_val(self._cmgr.open_cluster_group)

        self._cmgr.open_cluster.assert_called_once_with()
        self._cmgr.open_cluster_group.assert_called_once_with(
            self._FAKE_VM_NAME, cluster_handle=exp_clus_h)
        self._cmgr.open_cluster_node.assert_called_once_with(
            self._FAKE_HOST, cluster_handle=exp_clus_h)

        self._clusapi.move_cluster_group.assert_called_once_with(
            exp_clus_group_h, exp_clus_node_h, expected_migrate_flags,
            self._clusapi.get_property_list.return_value)

        mock_listener_cls.assert_called_once_with(exp_clus_h,
                                                  self._FAKE_VM_NAME)
        mock_listener = mock_listener_cls.return_value

        mock_wait_group.assert_called_once_with(
            mock_listener.__enter__.return_value,
            self._FAKE_VM_NAME, exp_clus_group_h,
            constants.CLUSTER_GROUP_ONLINE,
            mock.sentinel.timeout)

        if not wait_unexpected_exc:
            mock_validate_migr.assert_called_once_with(
                exp_clus_group_h,
                self._FAKE_VM_NAME,
                constants.CLUSTER_GROUP_ONLINE,
                self._FAKE_HOST)

    @mock.patch.object(clusterutils.ClusterUtils,
                       '_cancel_cluster_group_migration')
    @mock.patch.object(clusterutils.ClusterUtils,
                       '_wait_for_cluster_group_migration')
    @mock.patch.object(clusterutils.ClusterUtils,
                       '_validate_migration')
    @mock.patch.object(clusterutils,
                       '_ClusterGroupStateChangeListener')
    @ddt.data(True, False)
    def test_migrate_vm_timeout(self, finished_after_cancel,
                                mock_listener_cls,
                                mock_validate_migr,
                                mock_wait_group,
                                mock_cancel_migr):
        timeout_exc = exceptions.ClusterGroupMigrationTimeOut(
            group_name=self._FAKE_VM_NAME,
            time_elapsed=10)
        mock_wait_group.side_effect = timeout_exc
        mock_listener = self._cmgr_val(mock_listener_cls)
        mock_validate_migr.side_effect = (
            (None, ) if finished_after_cancel
            else exceptions.ClusterGroupMigrationFailed(
                group_name=self._FAKE_VM_NAME,
                expected_state=mock.sentinel.expected_state,
                expected_node=self._FAKE_HOST,
                group_state=mock.sentinel.expected_state,
                owner_node=mock.sentinel.other_host))

        migrate_args = (self._FAKE_VM_NAME,
                        self._FAKE_HOST,
                        self._clusterutils._LIVE_MIGRATION_TYPE,
                        mock.sentinel.exp_state,
                        mock.sentinel.timeout)

        if finished_after_cancel:
            self._clusterutils._migrate_vm(*migrate_args)
        else:
            self.assertRaises(exceptions.ClusterGroupMigrationTimeOut,
                              self._clusterutils._migrate_vm,
                              *migrate_args)

        exp_clus_group_h = self._cmgr_val(self._cmgr.open_cluster_group)
        mock_cancel_migr.assert_called_once_with(
            mock_listener, self._FAKE_VM_NAME, exp_clus_group_h,
            mock.sentinel.exp_state, mock.sentinel.timeout)
        mock_validate_migr.assert_called_once_with(exp_clus_group_h,
                                                   self._FAKE_VM_NAME,
                                                   mock.sentinel.exp_state,
                                                   self._FAKE_HOST)

    @ddt.data({},
              {'expected_state': constants.CLUSTER_GROUP_OFFLINE,
               'is_valid': False},
              {'expected_node': 'some_other_node',
               'is_valid': False})
    @ddt.unpack
    def test_validate_migration(
            self, expected_node=_FAKE_HOST,
            expected_state=constants.CLUSTER_GROUP_ONLINE,
            is_valid=True):
        group_state = dict(owner_node=self._FAKE_HOST.upper(),
                           state=constants.CLUSTER_GROUP_ONLINE)
        self._clusapi.get_cluster_group_state.return_value = group_state

        if is_valid:
            self._clusterutils._validate_migration(mock.sentinel.group_handle,
                                                   self._FAKE_VM_NAME,
                                                   expected_state,
                                                   expected_node)
        else:
            self.assertRaises(exceptions.ClusterGroupMigrationFailed,
                              self._clusterutils._validate_migration,
                              mock.sentinel.group_handle,
                              self._FAKE_VM_NAME,
                              expected_state,
                              expected_node)

        self._clusapi.get_cluster_group_state.assert_called_once_with(
            mock.sentinel.group_handle)

    @mock.patch.object(clusterutils.ClusterUtils,
                       '_cancel_cluster_group_migration')
    @mock.patch.object(clusterutils,
                       '_ClusterGroupStateChangeListener')
    def test_cancel_cluster_group_migration_public(self, mock_listener_cls,
                                                   mock_cancel_migr):

        exp_clus_h = self._cmgr_val(self._cmgr.open_cluster)
        exp_clus_group_h = self._cmgr_val(self._cmgr.open_cluster_group)

        mock_listener = mock_listener_cls.return_value
        mock_listener.__enter__.return_value = mock_listener

        self._clusterutils.cancel_cluster_group_migration(
            mock.sentinel.group_name,
            mock.sentinel.expected_state,
            mock.sentinel.timeout)

        self._cmgr.open_cluster.assert_called_once_with()
        self._cmgr.open_cluster_group.assert_called_once_with(
            mock.sentinel.group_name, cluster_handle=exp_clus_h)

        mock_listener.__enter__.assert_called_once_with()
        mock_listener_cls.assert_called_once_with(exp_clus_h,
                                                  mock.sentinel.group_name)
        mock_cancel_migr.assert_called_once_with(
            mock_listener,
            mock.sentinel.group_name,
            exp_clus_group_h,
            mock.sentinel.expected_state,
            mock.sentinel.timeout)

    @mock.patch.object(clusterutils.ClusterUtils,
                       '_get_cluster_group_state')
    @mock.patch.object(clusterutils.ClusterUtils,
                       '_is_migration_pending')
    @mock.patch.object(clusterutils.ClusterUtils,
                       '_wait_for_cluster_group_migration')
    @ddt.data({},
              {'cancel_exception': test_base.TestingException()},
              {'cancel_exception':
                  exceptions.Win32Exception(
                      error_code=w_const.INVALID_HANDLE_VALUE,
                      func_name=mock.sentinel.func_name,
                      error_message=mock.sentinel.error_message)},
              {'cancel_exception':
                  exceptions.Win32Exception(
                      error_code=w_const.ERROR_INVALID_STATE,
                      func_name=mock.sentinel.func_name,
                      error_message=mock.sentinel.error_message),
               'invalid_state_for_cancel': True},
              {'cancel_exception':
                  exceptions.Win32Exception(
                      error_code=w_const.ERROR_INVALID_STATE,
                      func_name=mock.sentinel.func_name,
                      error_message=mock.sentinel.error_message),
               'invalid_state_for_cancel': True,
               'cancel_still_pending': True},
              {'cancel_still_pending': True},
              {'cancel_still_pending': True,
               'cancel_wait_exception': test_base.TestingException()})
    @ddt.unpack
    def test_cancel_cluster_group_migration(self, mock_wait_migr,
                                            mock_is_migr_pending,
                                            mock_get_gr_state,
                                            cancel_still_pending=False,
                                            cancel_exception=None,
                                            invalid_state_for_cancel=False,
                                            cancel_wait_exception=None):
        expected_exception = None
        if cancel_wait_exception:
            expected_exception = exceptions.JobTerminateFailed()
        if (cancel_exception and (not invalid_state_for_cancel
                                  or cancel_still_pending)):
            expected_exception = cancel_exception

        mock_is_migr_pending.return_value = cancel_still_pending
        mock_get_gr_state.return_value = dict(
            state=mock.sentinel.state,
            status_info=mock.sentinel.status_info)

        self._clusapi.cancel_cluster_group_operation.side_effect = (
            cancel_exception or (not cancel_still_pending, ))
        mock_wait_migr.side_effect = cancel_wait_exception

        cancel_args = (mock.sentinel.listener,
                       mock.sentinel.group_name,
                       mock.sentinel.group_handle,
                       mock.sentinel.expected_state,
                       mock.sentinel.timeout)
        if expected_exception:
            self.assertRaises(
                expected_exception.__class__,
                self._clusterutils._cancel_cluster_group_migration,
                *cancel_args)
        else:
            self._clusterutils._cancel_cluster_group_migration(
                *cancel_args)

        self._clusapi.cancel_cluster_group_operation.assert_called_once_with(
            mock.sentinel.group_handle)

        if isinstance(cancel_exception, exceptions.Win32Exception):
            mock_get_gr_state.assert_called_once_with(
                mock.sentinel.group_handle)
            mock_is_migr_pending.assert_called_once_with(
                mock.sentinel.state,
                mock.sentinel.status_info,
                mock.sentinel.expected_state)
        if cancel_still_pending and not cancel_exception:
            mock_wait_migr.assert_called_once_with(
                mock.sentinel.listener,
                mock.sentinel.group_name,
                mock.sentinel.group_handle,
                mock.sentinel.expected_state,
                timeout=mock.sentinel.timeout)

    def test_is_migration_pending(self):
        self.assertTrue(
            self._clusterutils._is_migration_pending(
                group_state=constants.CLUSTER_GROUP_OFFLINE,
                group_status_info=0,
                expected_state=constants.CLUSTER_GROUP_ONLINE))
        self.assertTrue(
            self._clusterutils._is_migration_pending(
                group_state=constants.CLUSTER_GROUP_ONLINE,
                group_status_info=w_const.
                    CLUSGRP_STATUS_WAITING_IN_QUEUE_FOR_MOVE | 1,  # noqa
                expected_state=constants.CLUSTER_GROUP_ONLINE))
        self.assertFalse(
            self._clusterutils._is_migration_pending(
                group_state=constants.CLUSTER_GROUP_OFFLINE,
                group_status_info=0,
                expected_state=constants.CLUSTER_GROUP_OFFLINE))

    @mock.patch.object(clusterutils.ClusterUtils, '_is_migration_pending')
    @mock.patch.object(clusterutils.ClusterUtils, '_get_cluster_group_state')
    @mock.patch.object(clusterutils, 'time')
    def test_wait_for_clus_group_migr_timeout(self, mock_time,
                                              mock_get_gr_state,
                                              mock_is_migr_pending):
        exp_wait_iterations = 3
        mock_listener = mock.Mock()
        mock_time.time.side_effect = range(exp_wait_iterations + 2)
        timeout = 10

        state_info = dict(state=mock.sentinel.current_state,
                          status_info=mock.sentinel.status_info)

        events = [dict(status_info=mock.sentinel.migr_queued),
                  dict(state=mock.sentinel.pending_state),
                  queue.Empty]

        mock_get_gr_state.return_value = state_info
        mock_is_migr_pending.return_value = True
        mock_listener.get.side_effect = events

        self.assertRaises(
            exceptions.ClusterGroupMigrationTimeOut,
            self._clusterutils._wait_for_cluster_group_migration,
            mock_listener,
            mock.sentinel.group_name,
            mock.sentinel.group_handle,
            mock.sentinel.expected_state,
            timeout=timeout)

        mock_get_gr_state.assert_called_once_with(mock.sentinel.group_handle)

        exp_wait_times = [timeout - elapsed - 1
                          for elapsed in range(exp_wait_iterations)]
        mock_listener.get.assert_has_calls(
            [mock.call(wait_time) for wait_time in exp_wait_times])
        mock_is_migr_pending.assert_has_calls(
            [mock.call(mock.sentinel.current_state,
                       mock.sentinel.status_info,
                       mock.sentinel.expected_state),
             mock.call(mock.sentinel.current_state,
                       mock.sentinel.migr_queued,
                       mock.sentinel.expected_state),
             mock.call(mock.sentinel.pending_state,
                       mock.sentinel.migr_queued,
                       mock.sentinel.expected_state)])

    @mock.patch.object(clusterutils.ClusterUtils, '_is_migration_pending')
    @mock.patch.object(clusterutils.ClusterUtils, '_get_cluster_group_state')
    def test_wait_for_clus_group_migr_success(self, mock_get_gr_state,
                                              mock_is_migr_pending):
        mock_listener = mock.Mock()

        state_info = dict(state=mock.sentinel.current_state,
                          status_info=mock.sentinel.status_info)

        mock_get_gr_state.return_value = state_info
        mock_is_migr_pending.side_effect = [True, False]
        mock_listener.get.return_value = {}

        self._clusterutils._wait_for_cluster_group_migration(
            mock_listener,
            mock.sentinel.group_name,
            mock.sentinel.group_handle,
            mock.sentinel.expected_state,
            timeout=None)

        mock_listener.get.assert_called_once_with(None)

    @mock.patch.object(clusterutils.ClusterUtils, '_get_cluster_nodes')
    def get_cluster_node_name(self, mock_get_nodes):
        fake_node = dict(id=mock.sentinel.vm_id,
                         name=mock.sentinel.vm_name)
        mock_get_nodes.return_value([fake_node])

        self.assertEqual(
            mock.sentinel.vm_name,
            self._clusterutils.get_cluster_node_name(mock.sentinel.vm_id))
        self.assertRaises(
            exceptions.NotFound,
            self._clusterutils.get_cluster_node_name(mock.sentinel.missing_id))

    @mock.patch('ctypes.byref')
    def test_get_cluster_group_type(self, mock_byref):
        mock_byref.side_effect = lambda x: ('byref', x)
        self._clusapi.cluster_group_control.return_value = (
            mock.sentinel.buff, mock.sentinel.buff_sz)

        ret_val = self._clusterutils.get_cluster_group_type(
            mock.sentinel.group_name)
        self.assertEqual(
            self._clusapi.get_cluster_group_type.return_value,
            ret_val)

        self._cmgr.open_cluster_group.assert_called_once_with(
            mock.sentinel.group_name)
        self._clusapi.cluster_group_control.assert_called_once_with(
            self._cmgr_val(self._cmgr.open_cluster_group),
            w_const.CLUSCTL_GROUP_GET_RO_COMMON_PROPERTIES)
        self._clusapi.get_cluster_group_type.assert_called_once_with(
            mock_byref(mock.sentinel.buff), mock.sentinel.buff_sz)

    @mock.patch.object(clusterutils.ClusterUtils,
                       '_get_cluster_group_state')
    @mock.patch.object(clusterutils.ClusterUtils,
                       '_is_migration_queued')
    def test_get_cluster_group_state_info(self, mock_is_migr_queued,
                                          mock_get_gr_state):

        exp_clus_group_h = self._cmgr_val(self._cmgr.open_cluster_group)

        mock_get_gr_state.return_value = dict(
            state=mock.sentinel.state,
            status_info=mock.sentinel.status_info,
            owner_node=mock.sentinel.owner_node)

        sts_info = self._clusterutils.get_cluster_group_state_info(
            mock.sentinel.group_name)
        exp_sts_info = dict(state=mock.sentinel.state,
                            owner_node=mock.sentinel.owner_node,
                            migration_queued=mock_is_migr_queued.return_value)

        self.assertEqual(exp_sts_info, sts_info)

        self._cmgr.open_cluster_group.assert_called_once_with(
            mock.sentinel.group_name)

        mock_get_gr_state.assert_called_once_with(exp_clus_group_h)
        mock_is_migr_queued.assert_called_once_with(mock.sentinel.status_info)

    @mock.patch('ctypes.byref')
    def test_get_cluster_group_state(self, mock_byref):
        mock_byref.side_effect = lambda x: ('byref', x)

        state_info = dict(state=mock.sentinel.state,
                          owner_node=mock.sentinel.owner_node)
        self._clusapi.get_cluster_group_state.return_value = state_info

        self._clusapi.cluster_group_control.return_value = (
            mock.sentinel.buff, mock.sentinel.buff_sz)
        self._clusapi.get_cluster_group_status_info.return_value = (
            mock.sentinel.status_info)

        exp_state_info = state_info.copy()
        exp_state_info['status_info'] = mock.sentinel.status_info

        ret_val = self._clusterutils._get_cluster_group_state(
            mock.sentinel.group_handle)
        self.assertEqual(exp_state_info, ret_val)

        self._clusapi.get_cluster_group_state.assert_called_once_with(
            mock.sentinel.group_handle)
        self._clusapi.cluster_group_control.assert_called_once_with(
            mock.sentinel.group_handle,
            w_const.CLUSCTL_GROUP_GET_RO_COMMON_PROPERTIES)
        self._clusapi.get_cluster_group_status_info.assert_called_once_with(
            mock_byref(mock.sentinel.buff), mock.sentinel.buff_sz)

    @mock.patch.object(clusterutils, 'tpool')
    @mock.patch.object(clusterutils, 'patcher')
    def test_monitor_vm_failover_no_vm(self, mock_patcher, mock_tpool):
        mock_watcher = mock.MagicMock()
        fake_prev = mock.MagicMock(OwnerNode=self._FAKE_PREV_HOST)
        fake_wmi_object = mock.MagicMock(OwnerNode=self._FAKE_HOST,
                                         Name='Virtual Machine',
                                         previous=fake_prev)
        mock_tpool.execute.return_value = fake_wmi_object
        fake_callback = mock.MagicMock()

        self._clusterutils._monitor_vm_failover(mock_watcher,
                                                fake_callback,
                                                mock.sentinel.event_timeout_ms)

        mock_tpool.execute.assert_called_once_with(
            mock_watcher,
            mock.sentinel.event_timeout_ms)
        fake_callback.assert_not_called()

    @mock.patch.object(clusterutils, 'tpool')
    @mock.patch.object(clusterutils, 'patcher')
    def test_monitor_vm_failover(self, mock_patcher, mock_tpool):
        mock_watcher = mock.MagicMock()
        fake_prev = mock.MagicMock(OwnerNode=self._FAKE_PREV_HOST)
        fake_wmi_object = mock.MagicMock(OwnerNode=self._FAKE_HOST,
                                         Name=self._FAKE_RESOURCEGROUP_NAME,
                                         previous=fake_prev)
        mock_tpool.execute.return_value = fake_wmi_object
        fake_callback = mock.MagicMock()

        self._clusterutils._monitor_vm_failover(mock_watcher, fake_callback)

        mock_tpool.execute.assert_called_once_with(
            mock_watcher,
            self._clusterutils._WMI_EVENT_TIMEOUT_MS)
        fake_callback.assert_called_once_with(self._FAKE_VM_NAME,
                                              self._FAKE_PREV_HOST,
                                              self._FAKE_HOST)

    @mock.patch.object(clusterutils.ClusterUtils, '_get_failover_watcher')
    @mock.patch.object(clusterutils.ClusterUtils, '_monitor_vm_failover')
    @mock.patch.object(clusterutils, 'time')
    def test_get_vm_owner_change_listener(self, mock_time,
                                          mock_monitor, mock_get_watcher):
        mock_monitor.side_effect = [None, exceptions.OSWinException,
                                    KeyboardInterrupt]

        listener = self._clusterutils.get_vm_owner_change_listener()
        self.assertRaises(KeyboardInterrupt,
                          listener,
                          mock.sentinel.callback)

        mock_monitor.assert_has_calls(
            [mock.call(mock_get_watcher.return_value,
                       mock.sentinel.callback,
                       constants.DEFAULT_WMI_EVENT_TIMEOUT_MS)] * 3)
        mock_time.sleep.assert_called_once_with(
            constants.DEFAULT_WMI_EVENT_TIMEOUT_MS / 1000)

    @mock.patch.object(clusterutils._ClusterGroupOwnerChangeListener, 'get')
    @mock.patch.object(clusterutils.ClusterUtils, 'get_cluster_node_name')
    @mock.patch.object(clusterutils.ClusterUtils, 'get_cluster_group_type')
    @mock.patch.object(clusterutils, 'time')
    def test_get_vm_owner_change_listener_v2(self, mock_time, mock_get_type,
                                             mock_get_node_name,
                                             mock_get_event):
        mock_get_type.side_effect = [
            w_const.ClusGroupTypeVirtualMachine,
            mock.sentinel.other_type]
        mock_events = [mock.MagicMock(), mock.MagicMock()]
        mock_get_event.side_effect = (
            mock_events + [exceptions.OSWinException, KeyboardInterrupt])
        callback = mock.Mock()

        listener = self._clusterutils.get_vm_owner_change_listener_v2()
        self.assertRaises(KeyboardInterrupt,
                          listener,
                          callback)

        callback.assert_called_once_with(
            mock_events[0]['cluster_object_name'],
            mock_get_node_name.return_value)
        mock_get_node_name.assert_called_once_with(mock_events[0]['parent_id'])
        mock_get_type.assert_any_call(mock_events[0]['cluster_object_name'])
        mock_time.sleep.assert_called_once_with(
            constants.DEFAULT_WMI_EVENT_TIMEOUT_MS / 1000)


class ClusterEventListenerTestCase(test_base.OsWinBaseTestCase):
    @mock.patch.object(clusterutils._ClusterEventListener, '_setup')
    def setUp(self, mock_setup):
        super(ClusterEventListenerTestCase, self).setUp()

        self._setup_listener()

    def _setup_listener(self, stop_on_error=True):
        self._listener = clusterutils._ClusterEventListener(
            mock.sentinel.cluster_handle,
            stop_on_error=stop_on_error)

        self._listener._running = True
        self._listener._clusapi_utils = mock.Mock()
        self._clusapi = self._listener._clusapi_utils

    def test_get_notif_key_dw(self):
        fake_notif_key = 1
        notif_key_dw = self._listener._get_notif_key_dw(fake_notif_key)

        self.assertIsInstance(notif_key_dw, ctypes.c_ulong)
        self.assertEqual(fake_notif_key, notif_key_dw.value)
        self.assertEqual(notif_key_dw,
                         self._listener._get_notif_key_dw(fake_notif_key))

    @mock.patch.object(clusterutils._ClusterEventListener,
                       '_get_notif_key_dw')
    def test_add_filter(self, mock_get_notif_key):
        mock_get_notif_key.side_effect = (
            mock.sentinel.notif_key_dw,
            mock.sentinel.notif_key_dw_2)
        self._clusapi.create_cluster_notify_port_v2.return_value = (
            mock.sentinel.notif_port_h)

        self._listener._add_filter(mock.sentinel.filter,
                                   mock.sentinel.notif_key)
        self._listener._add_filter(mock.sentinel.filter_2,
                                   mock.sentinel.notif_key_2)

        self.assertEqual(mock.sentinel.notif_port_h,
                         self._listener._notif_port_h)
        mock_get_notif_key.assert_has_calls(
            [mock.call(mock.sentinel.notif_key),
             mock.call(mock.sentinel.notif_key_2)])
        self._clusapi.create_cluster_notify_port_v2.assert_has_calls(
            [mock.call(mock.sentinel.cluster_handle,
                       mock.sentinel.filter,
                       None,
                       mock.sentinel.notif_key_dw),
             mock.call(mock.sentinel.cluster_handle,
                       mock.sentinel.filter_2,
                       mock.sentinel.notif_port_h,
                       mock.sentinel.notif_key_dw_2)])

    @mock.patch.object(clusterutils._ClusterEventListener, '_add_filter')
    @mock.patch.object(clusapi_def, 'NOTIFY_FILTER_AND_TYPE')
    def test_setup_notif_port(self, mock_filter_struct_cls, mock_add_filter):
        notif_filter = dict(object_type=mock.sentinel.object_type,
                            filter_flags=mock.sentinel.filter_flags,
                            notif_key=mock.sentinel.notif_key)
        self._listener._notif_filters_list = [notif_filter]

        self._listener._setup_notif_port()

        mock_filter_struct_cls.assert_called_once_with(
            dwObjectType=mock.sentinel.object_type,
            FilterFlags=mock.sentinel.filter_flags)
        mock_add_filter.assert_called_once_with(
            mock_filter_struct_cls.return_value,
            mock.sentinel.notif_key)

    def test_signal_stopped(self):
        self._listener._signal_stopped()

        self.assertFalse(self._listener._running)
        self.assertIsNone(self._listener._event_queue.get(block=False))

    @mock.patch.object(clusterutils._ClusterEventListener,
                       '_signal_stopped')
    def test_stop(self, mock_signal_stopped):
        self._listener._notif_port_h = mock.sentinel.notif_port_h

        self._listener.stop()

        mock_signal_stopped.assert_called_once_with()
        self._clusapi.close_cluster_notify_port.assert_called_once_with(
            mock.sentinel.notif_port_h)

    @mock.patch.object(clusterutils._ClusterEventListener,
                       '_process_event')
    def test_listen(self, mock_process_event):
        events = [mock.sentinel.ignored_event, mock.sentinel.retrieved_event]
        self._clusapi.get_cluster_notify_v2.side_effect = events

        self._listener._notif_port_h = mock.sentinel.notif_port_h

        def fake_process_event(event):
            if event == mock.sentinel.ignored_event:
                return

            self._listener._running = False
            return mock.sentinel.processed_event

        mock_process_event.side_effect = fake_process_event

        self._listener._listen()

        processed_event = self._listener._event_queue.get(block=False)
        self.assertEqual(mock.sentinel.processed_event,
                         processed_event)
        self.assertTrue(self._listener._event_queue.empty())

        self._clusapi.get_cluster_notify_v2.assert_any_call(
            mock.sentinel.notif_port_h,
            timeout_ms=-1)

    def test_listen_exception(self):
        self._clusapi.get_cluster_notify_v2.side_effect = (
            test_base.TestingException)

        self._listener._listen()

        self.assertFalse(self._listener._running)

    @mock.patch.object(clusterutils._ClusterEventListener, '_setup')
    @mock.patch.object(clusterutils.time, 'sleep')
    def test_listen_ignore_exception(self, mock_sleep, mock_setup):
        self._setup_listener(stop_on_error=False)

        self._clusapi.get_cluster_notify_v2.side_effect = (
            test_base.TestingException,
            KeyboardInterrupt)

        self.assertRaises(KeyboardInterrupt, self._listener._listen)
        self.assertTrue(self._listener._running)
        mock_sleep.assert_called_once_with(
            self._listener._error_sleep_interval)

    def test_get_event(self):
        self._listener._event_queue = mock.Mock()

        event = self._listener.get(timeout=mock.sentinel.timeout)
        self.assertEqual(self._listener._event_queue.get.return_value, event)

        self._listener._event_queue.get.assert_called_once_with(
            timeout=mock.sentinel.timeout)

    def test_get_event_listener_stopped(self):
        self._listener._running = False
        self.assertRaises(exceptions.OSWinException,
                          self._listener.get,
                          timeout=1)

        def fake_get(block=True, timeout=0):
            self._listener._running = False
            return None

        self._listener._running = True
        self._listener._event_queue = mock.Mock(get=fake_get)

        self.assertRaises(exceptions.OSWinException,
                          self._listener.get,
                          timeout=1)

    @mock.patch.object(clusterutils._ClusterEventListener,
                       '_ensure_listener_running')
    @mock.patch.object(clusterutils._ClusterEventListener,
                       'stop')
    def test_context_manager(self, mock_stop, mock_ensure_running):
        with self._listener as l:
            self.assertIs(self._listener, l)
            mock_ensure_running.assert_called_once_with()

        mock_stop.assert_called_once_with()


class ClusterGroupStateChangeListenerTestCase(test_base.OsWinBaseTestCase):
    _FAKE_GROUP_NAME = 'fake_group_name'

    @mock.patch.object(clusterutils._ClusterEventListener, '_setup')
    def setUp(self, mock_setup):
        super(ClusterGroupStateChangeListenerTestCase, self).setUp()

        self._listener = clusterutils._ClusterGroupStateChangeListener(
            mock.sentinel.cluster_handle,
            self._FAKE_GROUP_NAME)

        self._listener._clusapi_utils = mock.Mock()
        self._clusapi = self._listener._clusapi_utils

    def _get_fake_event(self, **kwargs):
        event = dict(cluster_object_name=self._FAKE_GROUP_NAME.upper(),
                     object_type=mock.sentinel.object_type,
                     filter_flags=mock.sentinel.filter_flags,
                     buff=mock.sentinel.buff,
                     buff_sz=mock.sentinel.buff_sz)
        event.update(**kwargs)
        return event

    def _get_exp_processed_event(self, event, **kwargs):
        preserved_keys = ['cluster_object_name', 'object_type',
                          'filter_flags', 'notif_key']
        exp_proc_evt = {key: event[key] for key in preserved_keys}
        exp_proc_evt.update(**kwargs)
        return exp_proc_evt

    @mock.patch('ctypes.byref')
    def test_process_event_dropped(self, mock_byref):
        event = self._get_fake_event(cluster_object_name='other_group_name')
        self.assertIsNone(self._listener._process_event(event))

        event = self._get_fake_event(notif_key=2)
        self.assertIsNone(self._listener._process_event(event))

        notif_key = self._listener._NOTIF_KEY_GROUP_COMMON_PROP
        self._clusapi.get_cluster_group_status_info.side_effect = (
            exceptions.ClusterPropertyListEntryNotFound(
                property_name='fake_prop_name'))
        event = self._get_fake_event(notif_key=notif_key)
        self.assertIsNone(self._listener._process_event(event))

    def test_process_state_change_event(self):
        fake_state = constants.CLUSTER_GROUP_ONLINE
        event_buff = ctypes.c_ulong(fake_state)
        notif_key = self._listener._NOTIF_KEY_GROUP_STATE

        event = self._get_fake_event(notif_key=notif_key,
                                     buff=ctypes.byref(event_buff),
                                     buff_sz=ctypes.sizeof(event_buff))
        exp_proc_evt = self._get_exp_processed_event(
            event, state=fake_state)

        proc_evt = self._listener._process_event(event)
        self.assertEqual(exp_proc_evt, proc_evt)

    @mock.patch('ctypes.byref')
    def test_process_status_info_change_event(self, mock_byref):
        self._clusapi.get_cluster_group_status_info.return_value = (
            mock.sentinel.status_info)
        mock_byref.side_effect = lambda x: ('byref', x)
        notif_key = self._listener._NOTIF_KEY_GROUP_COMMON_PROP

        event = self._get_fake_event(notif_key=notif_key)
        exp_proc_evt = self._get_exp_processed_event(
            event, status_info=mock.sentinel.status_info)

        proc_evt = self._listener._process_event(event)
        self.assertEqual(exp_proc_evt, proc_evt)

        self._clusapi.get_cluster_group_status_info.assert_called_once_with(
            mock_byref(mock.sentinel.buff),
            mock.sentinel.buff_sz)



# Copyright 2017 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ctypes

from os_win.utils.winapi import wintypes

lib_handle = None


class VIRTUAL_STORAGE_TYPE(ctypes.Structure):
    _fields_ = [
        ('DeviceId', wintypes.DWORD),
        ('VendorId', wintypes.GUID)
    ]


PVIRTUAL_STORAGE_TYPE = ctypes.POINTER(VIRTUAL_STORAGE_TYPE)


class _RESIZE_VIRTUAL_DISK_PARAMETERS_V1(ctypes.Structure):
    _fields_ = [
        ('NewSize', wintypes.ULONGLONG)
    ]


# Only V1 is used, we avoid defining a union.
class RESIZE_VIRTUAL_DISK_PARAMETERS(ctypes.Structure):
    _fields_ = [
        ('Version', wintypes.DWORD),
        ('Version1', _RESIZE_VIRTUAL_DISK_PARAMETERS_V1)
    ]


PRESIZE_VIRTUAL_DISK_PARAMETERS = ctypes.POINTER(
    RESIZE_VIRTUAL_DISK_PARAMETERS)


class _OPEN_VIRTUAL_DISK_PARAMETERS_V1(ctypes.Structure):
    _fields_ = [
        ('RWDepth', wintypes.ULONG),
    ]


class _OPEN_VIRTUAL_DISK_PARAMETERS_V2(ctypes.Structure):
    _fields_ = [
        ('GetInfoOnly', wintypes.BOOL),
        ('ReadOnly', wintypes.BOOL),
        ('ResiliencyGuid', wintypes.GUID)
    ]


class _OPEN_VIRTUAL_DISK_PARAMETERS_U(ctypes.Union):
    _fields_ = [
        ('Version1', _OPEN_VIRTUAL_DISK_PARAMETERS_V1),
        ('Version2', _OPEN_VIRTUAL_DISK_PARAMETERS_V2)
    ]


class OPEN_VIRTUAL_DISK_PARAMETERS(ctypes.Structure):
    _anonymous_ = ['_parameters']
    _fields_ = [
        ('Version', wintypes.DWORD),
        ('_parameters', _OPEN_VIRTUAL_DISK_PARAMETERS_U)
    ]


POPEN_VIRTUAL_DISK_PARAMETERS = ctypes.POINTER(
    OPEN_VIRTUAL_DISK_PARAMETERS)


class _MERGE_VIRTUAL_DISK_PARAMETERS_V1(ctypes.Structure):
    _fields_ = [
        ('MergeDepth', wintypes.ULONG)
    ]


# Only V1 is used, we avoid defining a union.
class MERGE_VIRTUAL_DISK_PARAMETERS(ctypes.Structure):
    _fields_ = [
        ('Version', wintypes.DWORD),
        ('Version1', _MERGE_VIRTUAL_DISK_PARAMETERS_V1)
    ]


PMERGE_VIRTUAL_DISK_PARAMETERS = ctypes.POINTER(
    MERGE_VIRTUAL_DISK_PARAMETERS)


class _CREATE_VIRTUAL_DISK_PARAMETERS_V2(ctypes.Structure):
    _fields_ = [
        ('UniqueId', wintypes.GUID),
        ('MaximumSize', wintypes.ULONGLONG),
        ('BlockSizeInBytes', wintypes.ULONG),
        ('SectorSizeInBytes', wintypes.ULONG),
        ('PhysicalSectorSizeInBytes', wintypes.ULONG),
        ('ParentPath', wintypes.LPCWSTR),
        ('SourcePath', wintypes.LPCWSTR),
        ('OpenFlags', wintypes.DWORD),
        ('ParentVirtualStorageType', VIRTUAL_STORAGE_TYPE),
        ('SourceVirtualStorageType', VIRTUAL_STORAGE_TYPE),
        ('ResiliencyGuid', wintypes.GUID)
    ]


# Only V2 is used, we avoid defining a union.
class CREATE_VIRTUAL_DISK_PARAMETERS(ctypes.Structure):
    _fields_ = [
        ('Version', wintypes.DWORD),
        ('Version2', _CREATE_VIRTUAL_DISK_PARAMETERS_V2)
    ]


PCREATE_VIRTUAL_DISK_PARAMETERS = ctypes.POINTER(
    CREATE_VIRTUAL_DISK_PARAMETERS)


class _VHD_INFO_SIZE(ctypes.Structure):
    _fields_ = [
        ("VirtualSize", wintypes.ULARGE_INTEGER),
        ("PhysicalSize", wintypes.ULARGE_INTEGER),
        ("BlockSize", wintypes.ULONG),
        ("SectorSize", wintypes.ULONG)
    ]


class _VHD_INFO_PARENT_LOCATION(ctypes.Structure):
    _fields_ = [
        ('ParentResolved', wintypes.BOOL),
        ('ParentPath', wintypes.WCHAR * 512)
    ]


class _VHD_INFO_PHYSICAL_DISK(ctypes.Structure):
    _fields_ = [
        ("LogicalSectorSize", wintypes.ULONG),
        ("PhysicalSectorSize", wintypes.ULONG),
        ("IsRemote", wintypes.BOOL)
    ]


class _VHD_INFO(ctypes.Union):
    _fields_ = [
        ("Size", _VHD_INFO_SIZE),
        ("Identifier", wintypes.GUID),
        ("ParentLocation", _VHD_INFO_PARENT_LOCATION),
        ("ParentIdentifier", wintypes.GUID),
        ("ParentTimestamp", wintypes.ULONG),
        ("VirtualStorageType", VIRTUAL_STORAGE_TYPE),
        ("ProviderSubtype", wintypes.ULONG),
        ("Is4kAligned", wintypes.BOOL),
        ("PhysicalDisk", _VHD_INFO_PHYSICAL_DISK),
        ("VhdPhysicalSectorSize", wintypes.ULONG),
        ("SmallestSafeVirtualSize", wintypes.ULARGE_INTEGER),
        ("FragmentationPercentage", wintypes.ULONG)
    ]


class GET_VIRTUAL_DISK_INFO(ctypes.Structure):
    _anonymous_ = ['_vhdinfo']
    _fields_ = [
        ("Version", wintypes.UINT),
        ("_vhdinfo", _VHD_INFO)
    ]


PGET_VIRTUAL_DISK_INFO = ctypes.POINTER(GET_VIRTUAL_DISK_INFO)


# Only this version is used, we avoid defining a union.
class SET_VIRTUAL_DISK_INFO(ctypes.Structure):
    _fields_ = [
        ('Version', wintypes.DWORD),
        ('ParentFilePath', wintypes.LPCWSTR)
    ]


PSET_VIRTUAL_DISK_INFO = ctypes.POINTER(SET_VIRTUAL_DISK_INFO)


def register():
    global lib_handle
    lib_handle = ctypes.windll.virtdisk

    lib_handle.CreateVirtualDisk.argtypes = [
        PVIRTUAL_STORAGE_TYPE,
        wintypes.LPCWSTR,
        wintypes.INT,
        wintypes.PVOID,
        wintypes.INT,
        wintypes.ULONG,
        PCREATE_VIRTUAL_DISK_PARAMETERS,
        wintypes.LPOVERLAPPED,
        wintypes.PHANDLE
    ]
    lib_handle.CreateVirtualDisk.restype = wintypes.DWORD

    lib_handle.GetVirtualDiskInformation.argtypes = [
        wintypes.HANDLE,
        wintypes.PULONG,
        PGET_VIRTUAL_DISK_INFO,
        wintypes.PULONG
    ]
    lib_handle.GetVirtualDiskInformation.restype = wintypes.DWORD

    lib_handle.MergeVirtualDisk.argtypes = [
        wintypes.HANDLE,
        wintypes.INT,
        PMERGE_VIRTUAL_DISK_PARAMETERS,
        wintypes.LPOVERLAPPED
    ]
    lib_handle.MergeVirtualDisk.restype = wintypes.DWORD

    lib_handle.OpenVirtualDisk.argtypes = [
        PVIRTUAL_STORAGE_TYPE,
        wintypes.LPCWSTR,
        wintypes.INT,
        wintypes.INT,
        POPEN_VIRTUAL_DISK_PARAMETERS,
        wintypes.PHANDLE
    ]
    lib_handle.OpenVirtualDisk.restype = wintypes.DWORD

    lib_handle.ResizeVirtualDisk.argtypes = [
        wintypes.HANDLE,
        wintypes.INT,
        PRESIZE_VIRTUAL_DISK_PARAMETERS,
        wintypes.LPOVERLAPPED
    ]
    lib_handle.ResizeVirtualDisk.restype = wintypes.DWORD

    lib_handle.SetVirtualDiskInformation.argtypes = [
        wintypes.HANDLE,
        PSET_VIRTUAL_DISK_INFO
    ]
    lib_handle.SetVirtualDiskInformation.restype = wintypes.DWORD

    lib_handle.AttachVirtualDisk.argtypes = [
        wintypes.HANDLE,
        wintypes.PVOID,
        wintypes.INT,
        wintypes.ULONG,
        wintypes.PVOID,
        wintypes.LPOVERLAPPED
    ]
    lib_handle.AttachVirtualDisk.restype = wintypes.DWORD

    lib_handle.DetachVirtualDisk.argtypes = [
        wintypes.HANDLE,
        wintypes.INT,
        wintypes.ULONG
    ]
    lib_handle.DetachVirtualDisk.restype = wintypes.DWORD

    lib_handle.GetVirtualDiskPhysicalPath.argtypes = [
        wintypes.HANDLE,
        wintypes.PULONG,
        wintypes.PWSTR
    ]
    lib_handle.GetVirtualDiskPhysicalPath.restype = wintypes.DWORD



# Copyright 2014 Cloudbase Solutions SRL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unit tests for the Hyper-V utils factory.
"""

import inspect

import mock
from oslo_config import cfg
from oslo_utils import importutils

from os_win import exceptions
from os_win.tests.unit import test_base
from os_win.utils.compute import clusterutils
from os_win.utils.compute import livemigrationutils
from os_win.utils.compute import migrationutils
from os_win.utils.compute import rdpconsoleutils
from os_win.utils.compute import vmutils
from os_win.utils.dns import dnsutils
from os_win.utils import hostutils
from os_win.utils.network import networkutils
from os_win.utils import pathutils
from os_win.utils import processutils
from os_win.utils.storage import diskutils
from os_win.utils.storage.initiator import iscsi_utils
from os_win.utils.storage import smbutils
from os_win.utils.storage.virtdisk import vhdutils
from os_win import utilsfactory

CONF = cfg.CONF


class TestHyperVUtilsFactory(test_base.OsWinBaseTestCase):

    @mock.patch.object(utilsfactory.utils, 'get_windows_version')
    def test_get_class_unsupported_win_version(self, mock_get_windows_version):
        mock_get_windows_version.return_value = '5.2'
        self.assertRaises(exceptions.HyperVException, utilsfactory._get_class,
                          'hostutils')

    def test_get_class_unsupported_class_type(self):
        self.assertRaises(exceptions.HyperVException,
                          utilsfactory._get_class,
                          'invalid_class_type')

    @mock.patch.object(utilsfactory.utils, 'get_windows_version')
    def _check_get_class(self, mock_get_windows_version, expected_class,
                         class_type, windows_version='6.2', **kwargs):
        mock_get_windows_version.return_value = windows_version

        method = getattr(utilsfactory, 'get_%s' % class_type)
        instance = method(**kwargs)
        self.assertEqual(expected_class, type(instance))

        return instance

    def test_get_vmutils(self):
        instance = self._check_get_class(expected_class=vmutils.VMUtils,
                                         class_type='vmutils',
                                         host=mock.sentinel.host)
        self.assertEqual(mock.sentinel.host, instance._host)

    def test_get_vhdutils(self):
        self._check_get_class(expected_class=vhdutils.VHDUtils,
                              class_type='vhdutils')

    def test_get_networkutils(self):
        self._check_get_class(expected_class=networkutils.NetworkUtils,
                              class_type='networkutils')

    def test_get_networkutilsr2(self):
        self._check_get_class(expected_class=networkutils.NetworkUtilsR2,
                              class_type='networkutils',
                              windows_version='6.3')

    def test_get_hostutils(self):
        self._check_get_class(expected_class=hostutils.HostUtils,
                              class_type='hostutils')

    def test_get_pathutils(self):
        self._check_get_class(expected_class=pathutils.PathUtils,
                              class_type='pathutils')

    def test_get_livemigrationutils(self):
        self._check_get_class(
            expected_class=livemigrationutils.LiveMigrationUtils,
            class_type='livemigrationutils')

    @mock.patch.object(smbutils.SMBUtils, '__init__',
                       lambda *args, **kwargs: None)
    def test_get_smbutils(self):
        self._check_get_class(expected_class=smbutils.SMBUtils,
                              class_type='smbutils')

    def test_get_rdpconsoleutils(self):
        self._check_get_class(expected_class=rdpconsoleutils.RDPConsoleUtils,
                              class_type='rdpconsoleutils')

    def test_get_iscsi_initiator_utils(self):
        self._check_get_class(expected_class=iscsi_utils.ISCSIInitiatorUtils,
                              class_type='iscsi_initiator_utils')

    @mock.patch('os_win.utils.storage.initiator.fc_utils.FCUtils')
    def test_get_fc_utils(self, mock_cls_fcutils):
        self._check_get_class(
            expected_class=type(mock_cls_fcutils.return_value),
            class_type='fc_utils')

    def test_get_diskutils(self):
        self._check_get_class(
            expected_class=diskutils.DiskUtils,
            class_type='diskutils')

    def test_get_clusterutils(self):
        self._check_get_class(
            expected_class=clusterutils.ClusterUtils,
            class_type='clusterutils')

    def test_get_dnsutils(self):
        self._check_get_class(
            expected_class=dnsutils.DNSUtils,
            class_type='dnsutils')

    def test_get_migrationutils(self):
        self._check_get_class(
            expected_class=migrationutils.MigrationUtils,
            class_type='migrationutils')

    def test_get_processutils(self):
        self._check_get_class(
            expected_class=processutils.ProcessUtils,
            class_type='processutils')

    def test_utils_public_signatures(self):
        for module_name in utilsfactory.utils_map.keys():
            classes = utilsfactory.utils_map[module_name]
            class_names = list(classes.keys())
            if len(class_names) < 2:
                continue

            base_class_dict = classes[class_names[0]]
            base_class = importutils.import_object(base_class_dict['path'])
            for i in range(1, len(class_names)):
                tested_class_dict = classes[class_names[i]]
                tested_class = importutils.import_object(
                    tested_class_dict['path'])
                self.assertPublicAPISignatures(base_class, tested_class)
                self.assertPublicAPISignatures(tested_class, base_class)

    def assertPublicAPISignatures(self, baseinst, inst):
        def get_public_apis(inst):
            methods = {}
            for (name, value) in inspect.getmembers(inst, inspect.ismethod):
                if name.startswith("_"):
                    continue
                methods[name] = value
            return methods

        baseclass = baseinst.__class__.__name__
        basemethods = get_public_apis(baseinst)
        implmethods = get_public_apis(inst)

        extranames = [name for name in sorted(implmethods.keys()) if
                      name not in basemethods]
        self.assertEqual([], extranames,
                         "public methods not listed in class %s" % baseclass)

        for name in sorted(implmethods.keys()):
            baseargs = inspect.getargspec(basemethods[name])
            implargs = inspect.getargspec(implmethods[name])

            self.assertEqual(baseargs, implargs,
                             "%s args don't match class %s" %
                             (name, baseclass))



# Copyright 2016 Cloudbase Solutions Srl
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Utility class for metrics related operations.
Based on the "root/virtualization/v2" namespace available starting with
Hyper-V Server / Windows Server 2012.
"""

from oslo_log import log as logging

from os_win._i18n import _
from os_win import exceptions
from os_win.utils import _wqlutils
from os_win.utils import baseutils

LOG = logging.getLogger(__name__)


class MetricsUtils(baseutils.BaseUtilsVirt):

    _VIRTUAL_SYSTEM_TYPE_REALIZED = 'Microsoft:Hyper-V:System:Realized'
    _DVD_DISK_RES_SUB_TYPE = 'Microsoft:Hyper-V:Virtual CD/DVD Disk'
    _STORAGE_ALLOC_SETTING_DATA_CLASS = 'Msvm_StorageAllocationSettingData'
    _PROCESSOR_SETTING_DATA_CLASS = 'Msvm_ProcessorSettingData'
    _SYNTH_ETH_PORT_SET_DATA = 'Msvm_SyntheticEthernetPortSettingData'
    _PORT_ALLOC_SET_DATA = 'Msvm_EthernetPortAllocationSettingData'
    _PORT_ALLOC_ACL_SET_DATA = 'Msvm_EthernetSwitchPortAclSettingData'
    _BASE_METRICS_VALUE = 'Msvm_BaseMetricValue'

    _CPU_METRICS = 'Aggregated Average CPU Utilization'
    _MEMORY_METRICS = 'Aggregated Average Memory Utilization'
    _NET_IN_METRICS = 'Filtered Incoming Network Traffic'
    _NET_OUT_METRICS = 'Filtered Outgoing Network Traffic'
    # Disk metrics are supported from Hyper-V 2012 R2
    _DISK_RD_METRICS = 'Disk Data Read'
    _DISK_WR_METRICS = 'Disk Data Written'
    _DISK_LATENCY_METRICS = 'Average Disk Latency'
    _DISK_IOPS_METRICS = 'Average Normalized Disk Throughput'

    _METRICS_ENABLED = 2

    def __init__(self, host='.'):
        super(MetricsUtils, self).__init__(host)
        self._metrics_svc_obj = None
        self._metrics_defs_obj = {}

        # We need to avoid a circular dependency.
        from os_win import utilsfactory
        self._vmutils = utilsfactory.get_vmutils(host)

    @property
    def _metrics_svc(self):
        if not self._metrics_svc_obj:
            self._metrics_svc_obj = self._compat_conn.Msvm_MetricService()[0]
        return self._metrics_svc_obj

    @property
    def _metrics_defs(self):
        if not self._metrics_defs_obj:
            self._cache_metrics_defs()
        return self._metrics_defs_obj

    def _cache_metrics_defs(self):
        for metrics_def in self._conn.CIM_BaseMetricDefinition():
            self._metrics_defs_obj[metrics_def.ElementName] = metrics_def

    def enable_vm_metrics_collection(self, vm_name):
        vm = self._get_vm(vm_name)
        disks = self._get_vm_resources(vm_name,
                                       self._STORAGE_ALLOC_SETTING_DATA_CLASS)
        filtered_disks = [d for d in disks if
                          d.ResourceSubType != self._DVD_DISK_RES_SUB_TYPE]

        # enable metrics for disk.
        for disk in filtered_disks:
            self._enable_metrics(disk)

        metrics_names = [self._CPU_METRICS, self._MEMORY_METRICS]
        self._enable_metrics(vm, metrics_names)

    def enable_disk_metrics_collection(self, attached_disk_path=None,
                                       is_physical=False,
                                       serial=None):
        disk = self._vmutils._get_mounted_disk_resource_from_path(
            attached_disk_path, is_physical=is_physical, serial=serial)
        self._enable_metrics(disk)

    def enable_port_metrics_collection(self, switch_port_name):
        port = self._get_switch_port(switch_port_name)
        metrics_names = [self._NET_IN_METRICS, self._NET_OUT_METRICS]
        self._enable_metrics(port, metrics_names)

    def _enable_metrics(self, element, metrics_names=None):
        if not metrics_names:
            definition_paths = [None]
        else:
            definition_paths = []
            for metrics_name in metrics_names:
                metrics_def = self._metrics_defs.get(metrics_name)
                if not metrics_def:
                    LOG.warning("Metric not found: %s", metrics_name)
                    continue
                definition_paths.append(metrics_def.path_())

        element_path = element.path_()
        for definition_path in definition_paths:
            ret_val = self._metrics_svc.ControlMetrics(
                Subject=element_path,
                Definition=definition_path,
                MetricCollectionEnabled=self._METRICS_ENABLED)[0]
            if ret_val:
                err_msg = _("Failed to enable metrics for resource "
                            "%(resource_name)s. "
                            "Return code: %(ret_val)s.") % dict(
                                resource_name=element.ElementName,
                                ret_val=ret_val)
                raise exceptions.OSWinException(err_msg)

    def get_cpu_metrics(self, vm_name):
        vm = self._get_vm(vm_name)
        cpu_sd = self._get_vm_resources(vm_name,
                                        self._PROCESSOR_SETTING_DATA_CLASS)[0]
        cpu_metrics_def = self._metrics_defs[self._CPU_METRICS]
        cpu_metrics_aggr = self._get_metrics(vm, cpu_metrics_def)

        cpu_used = 0
        if cpu_metrics_aggr:
            cpu_used = int(cpu_metrics_aggr[0].MetricValue)

        return (cpu_used,
                int(cpu_sd.VirtualQuantity),
                int(vm.OnTimeInMilliseconds))

    def get_memory_metrics(self, vm_name):
        vm = self._get_vm(vm_name)
        memory_def = self._metrics_defs[self._MEMORY_METRICS]
        metrics_memory = self._get_metrics(vm, memory_def)
        memory_usage = 0
        if metrics_memory:
            memory_usage = int(metrics_memory[0].MetricValue)
        return memory_usage

    def get_vnic_metrics(self, vm_name):
        ports = self._get_vm_resources(vm_name, self._PORT_ALLOC_SET_DATA)
        vnics = self._get_vm_resources(vm_name, self._SYNTH_ETH_PORT_SET_DATA)

        metrics_def_in = self._metrics_defs[self._NET_IN_METRICS]
        metrics_def_out = self._metrics_defs[self._NET_OUT_METRICS]

        for port in ports:
            vnic = [v for v in vnics if port.Parent == v.path_()][0]
            port_acls = _wqlutils.get_element_associated_class(
                self._conn, self._PORT_ALLOC_ACL_SET_DATA,
                element_instance_id=port.InstanceID)

            metrics_value_instances = self._get_metrics_value_instances(
                port_acls, self._BASE_METRICS_VALUE)
            metrics_values = self._sum_metrics_values_by_defs(
                metrics_value_instances, [metrics_def_in, metrics_def_out])

            yield {
                'rx_mb': metrics_values[0],
                'tx_mb': metrics_values[1],
                'element_name': vnic.ElementName,
                'address': vnic.Address
            }

    def get_disk_metrics(self, vm_name):
        metrics_def_r = self._metrics_defs[self._DISK_RD_METRICS]
        metrics_def_w = self._metrics_defs[self._DISK_WR_METRICS]

        disks = self._get_vm_resources(vm_name,
                                       self._STORAGE_ALLOC_SETTING_DATA_CLASS)
        for disk in disks:
            metrics_values = self._get_metrics_values(
                disk, [metrics_def_r, metrics_def_w])

            yield {
                # Values are in megabytes
                'read_mb': metrics_values[0],
                'write_mb': metrics_values[1],
                'instance_id': disk.InstanceID,
                'host_resource': disk.HostResource[0]
            }

    def get_disk_latency_metrics(self, vm_name):
        metrics_latency_def = self._metrics_defs[self._DISK_LATENCY_METRICS]

        disks = self._get_vm_resources(vm_name,
                                       self._STORAGE_ALLOC_SETTING_DATA_CLASS)
        for disk in disks:
            metrics_values = self._get_metrics_values(
                disk, [metrics_latency_def])

            yield {
                'disk_latency': metrics_values[0],
                'instance_id': disk.InstanceID,
            }

    def get_disk_iops_count(self, vm_name):
        metrics_def_iops = self._metrics_defs[self._DISK_IOPS_METRICS]

        disks = self._get_vm_resources(vm_name,
                                       self._STORAGE_ALLOC_SETTING_DATA_CLASS)
        for disk in disks:
            metrics_values = self._get_metrics_values(
                disk, [metrics_def_iops])

            yield {
                'iops_count': metrics_values[0],
                'instance_id': disk.InstanceID,
            }

    @staticmethod
    def _sum_metrics_values(metrics):
        return sum([int(metric.MetricValue) for metric in metrics])

    def _sum_metrics_values_by_defs(self, element_metrics, metrics_defs):
        metrics_values = []
        for metrics_def in metrics_defs:
            if metrics_def:
                metrics = self._filter_metrics(element_metrics, metrics_def)
                metrics_values.append(self._sum_metrics_values(metrics))
            else:
                # In case the metric is not defined on this host
                metrics_values.append(0)
        return metrics_values

    def _get_metrics_value_instances(self, elements, result_class):
        instances = []
        for el in elements:
            # NOTE(abalutoiu): Msvm_MetricForME is the association between
            # an element and all the metric values maintained for it.
            el_metric = [
                x.Dependent for x in self._conn.Msvm_MetricForME(
                    Antecedent=el.path_())]
            el_metric = [
                x for x in el_metric if x.path().Class == result_class]
            if el_metric:
                instances.append(el_metric[0])

        return instances

    def _get_metrics_values(self, element, metrics_defs):
        element_metrics = [
            x.Dependent for x in self._conn.Msvm_MetricForME(
                Antecedent=element.path_())]
        return self._sum_metrics_values_by_defs(element_metrics, metrics_defs)

    def _get_metrics(self, element, metrics_def):
        metrics = [
            x.Dependent for x in self._conn.Msvm_MetricForME(
                Antecedent=element.path_())]
        return self._filter_metrics(metrics, metrics_def)

    @staticmethod
    def _filter_metrics(all_metrics, metrics_def):
        return [v for v in all_metrics if
                v.MetricDefinitionId == metrics_def.Id]

    def _get_vm_resources(self, vm_name, resource_class):
        setting_data = self._get_vm_setting_data(vm_name)
        return _wqlutils.get_element_associated_class(
            self._conn, resource_class,
            element_instance_id=setting_data.InstanceID)

    def _get_vm(self, vm_name):
        vms = self._conn.Msvm_ComputerSystem(ElementName=vm_name)
        return self._unique_result(vms, vm_name)

    def _get_switch_port(self, port_name):
        ports = self._conn.Msvm_EthernetPortAllocationSettingData(
            ElementName=port_name)
        return self._unique_result(ports, port_name)

    def _get_vm_setting_data(self, vm_name):
        vssds = self._conn.Msvm_VirtualSystemSettingData(
            ElementName=vm_name)
        return self._unique_result(vssds, vm_name)

    @staticmethod
    def _unique_result(objects, resource_name):
        n = len(objects)
        if n == 0:
            raise exceptions.NotFound(resource=resource_name)
        elif n > 1:
            raise exceptions.OSWinException(
                _('Duplicate resource name found: %s') % resource_name)
        else:
            return objects[0]



#  Copyright 2013 Cloudbase Solutions Srl
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ctypes
import os

import ddt
import mock
import six

from os_win import constants
from os_win import exceptions
from os_win.tests.unit import test_base
from os_win.utils.storage.virtdisk import vhdutils
from os_win.utils.winapi import constants as w_const
from os_win.utils.winapi import wintypes


@ddt.ddt
class VHDUtilsTestCase(test_base.BaseTestCase):
    """Unit tests for the Hyper-V VHDUtils class."""

    _autospec_classes = [
        vhdutils.win32utils.Win32Utils,
    ]

    def setUp(self):
        super(VHDUtilsTestCase, self).setUp()
        self._setup_lib_mocks()

        self._fake_vst_struct = self._vdisk_struct.VIRTUAL_STORAGE_TYPE

        self._vhdutils = vhdutils.VHDUtils()

        self._mock_close = self._vhdutils._win32_utils.close_handle
        self._mock_run = self._vhdutils._win32_utils.run_and_check_output
        self._run_args = self._vhdutils._virtdisk_run_args

        self.addCleanup(mock.patch.stopall)

    def _setup_lib_mocks(self):
        self._vdisk_struct = mock.Mock()
        self._ctypes = mock.Mock()
        # This is used in order to easily make assertions on the variables
        # passed by reference.
        self._ctypes.byref = lambda x: (x, "byref")
        self._ctypes.c_wchar_p = lambda x: (x, "c_wchar_p")
        self._ctypes.c_ulong = lambda x: (x, "c_ulong")

        self._ctypes_patcher = mock.patch.object(
            vhdutils, 'ctypes', self._ctypes)
        self._ctypes_patcher.start()

        mock.patch.multiple(vhdutils,
                            kernel32=mock.DEFAULT,
                            wintypes=mock.DEFAULT, virtdisk=mock.DEFAULT,
                            vdisk_struct=self._vdisk_struct,
                            create=True).start()

    def _test_run_and_check_output(self, raised_exc=None):
        self._mock_run.side_effect = raised_exc(
            func_name='fake_func_name',
            error_code='fake_error_code',
            error_message='fake_error_message') if raised_exc else None

        if raised_exc:
            self.assertRaises(
                raised_exc,
                self._vhdutils._run_and_check_output,
                mock.sentinel.func,
                mock.sentinel.arg,
                cleanup_handle=mock.sentinel.handle)
        else:
            ret_val = self._vhdutils._run_and_check_output(
                mock.sentinel.func,
                mock.sentinel.arg,
                cleanup_handle=mock.sentinel.handle)
            self.assertEqual(self._mock_run.return_value, ret_val)

        self._mock_run.assert_called_once_with(
            mock.sentinel.func, mock.sentinel.arg, **self._run_args)
        self._mock_close.assert_called_once_with(mock.sentinel.handle)

    def test_run_and_check_output(self):
        self._test_run_and_check_output()

    def test_run_and_check_output_raising_error(self):
        self._test_run_and_check_output(
            raised_exc=exceptions.VHDWin32APIException)

    @mock.patch.object(vhdutils.VHDUtils, '_get_vhd_device_id')
    def test_open(self, mock_get_dev_id):
        fake_vst = self._fake_vst_struct.return_value

        mock_get_dev_id.return_value = mock.sentinel.device_id

        handle = self._vhdutils._open(
            vhd_path=mock.sentinel.vhd_path,
            open_flag=mock.sentinel.open_flag,
            open_access_mask=mock.sentinel.access_mask,
            open_params=mock.sentinel.open_params)

        self.assertEqual(vhdutils.wintypes.HANDLE.return_value, handle)
        self._fake_vst_struct.assert_called_once_with(
            DeviceId=mock.sentinel.device_id,
            VendorId=w_const.VIRTUAL_STORAGE_TYPE_VENDOR_MICROSOFT)

        self._mock_run.assert_called_once_with(
            vhdutils.virtdisk.OpenVirtualDisk,
            self._ctypes.byref(fake_vst),
            self._ctypes.c_wchar_p(mock.sentinel.vhd_path),
            mock.sentinel.access_mask,
            mock.sentinel.open_flag,
            mock.sentinel.open_params,
            self._ctypes.byref(vhdutils.wintypes.HANDLE.return_value),
            **self._run_args)

    def test_close(self):
        self._vhdutils.close(mock.sentinel.handle)
        self._mock_close.assert_called_once_with(
            mock.sentinel.handle)

    @mock.patch.object(vhdutils.VHDUtils, '_get_vhd_device_id')
    def _test_create_vhd(self, mock_get_dev_id, new_vhd_type):
        create_params_struct = (
            self._vdisk_struct.CREATE_VIRTUAL_DISK_PARAMETERS)
        mock_handle = vhdutils.wintypes.HANDLE.return_value

        fake_vst = self._fake_vst_struct.return_value
        fake_create_params = create_params_struct.return_value

        expected_create_vhd_flag = (
            vhdutils.CREATE_VIRTUAL_DISK_FLAGS.get(new_vhd_type, 0))

        self._vhdutils.create_vhd(
            new_vhd_path=mock.sentinel.new_vhd_path,
            new_vhd_type=new_vhd_type,
            src_path=mock.sentinel.src_path,
            max_internal_size=mock.sentinel.max_internal_size,
            parent_path=mock.sentinel.parent_path)

        self._fake_vst_struct.assert_called_once_with(
            DeviceId=mock_get_dev_id.return_value,
            VendorId=w_const.VIRTUAL_STORAGE_TYPE_VENDOR_MICROSOFT)

        self.assertEqual(w_const.CREATE_VIRTUAL_DISK_VERSION_2,
                         fake_create_params.Version)
        self.assertEqual(mock.sentinel.max_internal_size,
                         fake_create_params.Version2.MaximumSize)
        self.assertEqual(mock.sentinel.parent_path,
                         fake_create_params.Version2.ParentPath)
        self.assertEqual(mock.sentinel.src_path,
                         fake_create_params.Version2.SourcePath)
        self.assertEqual(
            vhdutils.VIRTUAL_DISK_DEFAULT_PHYS_SECTOR_SIZE,
            fake_create_params.Version2.PhysicalSectorSizeInBytes)
        self.assertEqual(
            w_const.CREATE_VHD_PARAMS_DEFAULT_BLOCK_SIZE,
            fake_create_params.Version2.BlockSizeInBytes)
        self.assertEqual(
            vhdutils.VIRTUAL_DISK_DEFAULT_SECTOR_SIZE,
            fake_create_params.Version2.SectorSizeInBytes)

        self._mock_run.assert_called_once_with(
            vhdutils.virtdisk.CreateVirtualDisk,
            self._ctypes.byref(fake_vst),
            self._ctypes.c_wchar_p(mock.sentinel.new_vhd_path),
            0,
            None,
            expected_create_vhd_flag,
            0,
            self._ctypes.byref(fake_create_params),
            None,
            self._ctypes.byref(mock_handle),
            **self._run_args)

        self._mock_close.assert_called_once_with(mock_handle)

    def test_create_dynamic_vhd(self):
        self._test_create_vhd(new_vhd_type=constants.VHD_TYPE_DYNAMIC)

    def test_create_fixed_vhd(self):
        self._test_create_vhd(new_vhd_type=constants.VHD_TYPE_FIXED)

    @mock.patch.object(vhdutils.VHDUtils, 'create_vhd')
    def test_create_dynamic_vhd_helper(self, mock_create_vhd):
        self._vhdutils.create_dynamic_vhd(mock.sentinel.path,
                                          mock.sentinel.size)

        mock_create_vhd.assert_called_once_with(
            mock.sentinel.path,
            constants.VHD_TYPE_DYNAMIC,
            max_internal_size=mock.sentinel.size)

    @mock.patch.object(vhdutils.VHDUtils, 'create_vhd')
    def test_create_differencing_vhd_helper(self, mock_create_vhd):
        self._vhdutils.create_differencing_vhd(mock.sentinel.path,
                                               mock.sentinel.parent_path)

        mock_create_vhd.assert_called_once_with(
            mock.sentinel.path,
            constants.VHD_TYPE_DIFFERENCING,
            parent_path=mock.sentinel.parent_path)

    @mock.patch.object(vhdutils.VHDUtils, 'create_vhd')
    def test_convert_vhd(self, mock_create_vhd):
        self._vhdutils.convert_vhd(mock.sentinel.src,
                                   mock.sentinel.dest,
                                   mock.sentinel.vhd_type)

        mock_create_vhd.assert_called_once_with(
            mock.sentinel.dest,
            mock.sentinel.vhd_type,
            src_path=mock.sentinel.src)

    def test_get_vhd_format_found_by_ext(self):
        fake_vhd_path = 'C:\\test.vhd'

        ret_val = self._vhdutils.get_vhd_format(fake_vhd_path)

        self.assertEqual(constants.DISK_FORMAT_VHD, ret_val)

    @mock.patch.object(vhdutils.VHDUtils, '_get_vhd_format_by_signature')
    @mock.patch('os.path.exists')
    def _test_vhd_format_unrecognized_ext(self, mock_exists,
                                          mock_get_vhd_fmt_by_sign,
                                          signature_available=False):
        mock_exists.return_value = True
        fake_vhd_path = 'C:\\test_vhd'
        mock_get_vhd_fmt_by_sign.return_value = (
            constants.DISK_FORMAT_VHD if signature_available else None)

        if signature_available:
            ret_val = self._vhdutils.get_vhd_format(fake_vhd_path)
            self.assertEqual(constants.DISK_FORMAT_VHD, ret_val)
        else:
            self.assertRaises(exceptions.VHDException,
                              self._vhdutils.get_vhd_format,
                              fake_vhd_path)

    def test_get_vhd_format_unrecognised_ext_unavailable_signature(self):
        self._test_vhd_format_unrecognized_ext()

    def test_get_vhd_format_unrecognised_ext_available_signature(self):
        self._test_vhd_format_unrecognized_ext(signature_available=True)

    @mock.patch.object(vhdutils.VHDUtils, 'get_vhd_format')
    def test_get_vhd_device_id(self, mock_get_vhd_fmt):
        mock_get_vhd_fmt.return_value = constants.DISK_FORMAT_VHD

        dev_id = self._vhdutils._get_vhd_device_id(mock.sentinel.vhd_path)

        mock_get_vhd_fmt.assert_called_once_with(mock.sentinel.vhd_path)
        self.assertEqual(w_const.VIRTUAL_STORAGE_TYPE_DEVICE_VHD,
                         dev_id)

    def _mock_open(self, read_data=None, curr_f_pos=0):
        mock_open = mock.mock_open()
        mock.patch.object(vhdutils, 'open', mock_open,
                          create=True).start()

        f = mock_open.return_value
        f.read.side_effect = read_data
        f.tell.return_value = curr_f_pos

        return mock_open

    def test_get_vhd_format_by_sig_vhdx(self):
        read_data = (vhdutils.VHDX_SIGNATURE, )
        self._mock_open(read_data=read_data)

        fmt = self._vhdutils._get_vhd_format_by_signature(
            mock.sentinel.vhd_path)

        self.assertEqual(constants.DISK_FORMAT_VHDX, fmt)

    def test_get_vhd_format_by_sig_vhd(self):
        read_data = ('notthesig', vhdutils.VHD_SIGNATURE)
        mock_open = self._mock_open(read_data=read_data, curr_f_pos=1024)

        fmt = self._vhdutils._get_vhd_format_by_signature(
            mock.sentinel.vhd_path)

        self.assertEqual(constants.DISK_FORMAT_VHD, fmt)
        mock_open.return_value.seek.assert_has_calls([mock.call(0, 2),
                                                      mock.call(-512, 2)])

    def test_get_vhd_format_by_sig_invalid_format(self):
        self._mock_open(read_data='notthesig', curr_f_pos=1024)

        fmt = self._vhdutils._get_vhd_format_by_signature(
            mock.sentinel.vhd_path)

        self.assertIsNone(fmt)

    def test_get_vhd_format_by_sig_zero_length_file(self):
        mock_open = self._mock_open(read_data=('', ''))

        fmt = self._vhdutils._get_vhd_format_by_signature(
            mock.sentinel.vhd_path)

        self.assertIsNone(fmt)
        mock_open.return_value.seek.assert_called_once_with(0, 2)

    @mock.patch.object(vhdutils.VHDUtils, '_open')
    @mock.patch.object(vhdutils.VHDUtils, '_get_vhd_info_member')
    def test_get_vhd_info(self, mock_get_vhd_info_member,
                          mock_open):
        fake_info_member = w_const.GET_VIRTUAL_DISK_INFO_SIZE
        fake_vhd_info = {'VirtualSize': mock.sentinel.virtual_size}

        mock_open.return_value = mock.sentinel.handle
        mock_get_vhd_info_member.return_value = fake_vhd_info

        expected_open_flag = w_const.OPEN_VIRTUAL_DISK_FLAG_NO_PARENTS
        expected_access_mask = (w_const.VIRTUAL_DISK_ACCESS_GET_INFO |
                                w_const.VIRTUAL_DISK_ACCESS_DETACH)

        ret_val = self._vhdutils.get_vhd_info(mock.sentinel.vhd_path,
                                              [fake_info_member])

        self.assertEqual(fake_vhd_info, ret_val)
        mock_open.assert_called_once_with(
            mock.sentinel.vhd_path,
            open_flag=expected_open_flag,
            open_access_mask=expected_access_mask)
        self._vhdutils._get_vhd_info_member.assert_called_once_with(
            mock.sentinel.handle,
            fake_info_member)
        self._mock_close.assert_called_once_with(mock.sentinel.handle)

    @mock.patch.object(vhdutils.VHDUtils, '_parse_vhd_info')
    def test_get_vhd_info_member(self, mock_parse_vhd_info):
        get_vd_info_struct = (
            self._vdisk_struct.GET_VIRTUAL_DISK_INFO)
        fake_params = get_vd_info_struct.return_value
        fake_info_size = self._ctypes.sizeof.return_value

        info_member = w_const.GET_VIRTUAL_DISK_INFO_PARENT_LOCATION

        vhd_info = self._vhdutils._get_vhd_info_member(
            mock.sentinel.vhd_path,
            info_member)

        self._mock_run.assert_called_once_with(
            vhdutils.virtdisk.GetVirtualDiskInformation,
            mock.sentinel.vhd_path,
            self._ctypes.byref(
                self._ctypes.c_ulong(fake_info_size)),
            self._ctypes.byref(fake_params), None,
            ignored_error_codes=[w_const.ERROR_VHD_INVALID_TYPE],
            **self._run_args)

        self.assertEqual(mock_parse_vhd_info.return_value, vhd_info)
        mock_parse_vhd_info.assert_called_once_with(fake_params,
                                                    info_member)

    def test_parse_vhd_info(self):
        fake_info_member = w_const.GET_VIRTUAL_DISK_INFO_SIZE
        fake_info = mock.Mock()
        fake_info.Size._fields_ = [
            ("VirtualSize", vhdutils.wintypes.ULARGE_INTEGER),
            ("PhysicalSize", vhdutils.wintypes.ULARGE_INTEGER)]
        fake_info.Size.VirtualSize = mock.sentinel.virt_size
        fake_info.Size.PhysicalSize = mock.sentinel.phys_size

        ret_val = self._vhdutils._parse_vhd_info(fake_info,
                                                 fake_info_member)
        expected = {'VirtualSize': mock.sentinel.virt_size,
                    'PhysicalSize': mock.sentinel.phys_size}

        self.assertEqual(expected, ret_val)

    def test_parse_vhd_provider_subtype_member(self):
        fake_info_member = w_const.GET_VIRTUAL_DISK_INFO_PROVIDER_SUBTYPE
        fake_info = mock.Mock()
        fake_info.ProviderSubtype = mock.sentinel.provider_subtype

        ret_val = self._vhdutils._parse_vhd_info(fake_info, fake_info_member)
        expected = {'ProviderSubtype': mock.sentinel.provider_subtype}

        self.assertEqual(expected, ret_val)

    @mock.patch.object(vhdutils.VHDUtils, 'get_vhd_info')
    def test_get_vhd_size(self, mock_get_vhd_info):
        ret_val = self._vhdutils.get_vhd_size(mock.sentinel.vhd_path)

        self.assertEqual(mock_get_vhd_info.return_value, ret_val)
        mock_get_vhd_info.assert_called_once_with(
            mock.sentinel.vhd_path,
            [w_const.GET_VIRTUAL_DISK_INFO_SIZE])

    @mock.patch.object(vhdutils.VHDUtils, 'get_vhd_info')
    def test_get_vhd_parent_path(self, mock_get_vhd_info):
        mock_get_vhd_info.return_value = {
            'ParentPath': mock.sentinel.parent_path}

        ret_val = self._vhdutils.get_vhd_parent_path(mock.sentinel.vhd_path)

        self.assertEqual(mock.sentinel.parent_path, ret_val)
        mock_get_vhd_info.assert_called_once_with(
            mock.sentinel.vhd_path,
            [w_const.GET_VIRTUAL_DISK_INFO_PARENT_LOCATION])

    @mock.patch.object(vhdutils.VHDUtils, 'get_vhd_info')
    def test_get_vhd_type(self, mock_get_vhd_info):
        mock_get_vhd_info.return_value = {
            'ProviderSubtype': mock.sentinel.provider_subtype}

        ret_val = self._vhdutils.get_vhd_type(mock.sentinel.vhd_path)

        self.assertEqual(mock.sentinel.provider_subtype, ret_val)
        mock_get_vhd_info.assert_called_once_with(
            mock.sentinel.vhd_path,
            [w_const.GET_VIRTUAL_DISK_INFO_PROVIDER_SUBTYPE])

    @mock.patch.object(vhdutils.VHDUtils, '_open')
    @mock.patch('os.remove')
    def test_merge_vhd(self, mock_remove, mock_open):
        open_params_struct = (
            self._vdisk_struct.OPEN_VIRTUAL_DISK_PARAMETERS)
        merge_params_struct = (
            self._vdisk_struct.MERGE_VIRTUAL_DISK_PARAMETERS)

        fake_open_params = open_params_struct.return_value
        fake_merge_params = merge_params_struct.return_value
        mock_open.return_value = mock.sentinel.handle

        self._vhdutils.merge_vhd(mock.sentinel.vhd_path)

        self.assertEqual(w_const.OPEN_VIRTUAL_DISK_VERSION_1,
                         fake_open_params.Version)
        self.assertEqual(2,
                         fake_open_params.Version1.RWDepth)

        mock_open.assert_called_once_with(
            mock.sentinel.vhd_path,
            open_params=self._ctypes.byref(fake_open_params))

        self.assertEqual(w_const.MERGE_VIRTUAL_DISK_VERSION_1,
                         fake_merge_params.Version)
        self.assertEqual(1,
                         fake_merge_params.Version1.MergeDepth)

        self._mock_run.assert_called_once_with(
            vhdutils.virtdisk.MergeVirtualDisk,
            mock.sentinel.handle,
            0,
            self._ctypes.byref(fake_merge_params),
            None,
            **self._run_args)
        mock_remove.assert_called_once_with(
            mock.sentinel.vhd_path)
        self._mock_close.assert_called_once_with(mock.sentinel.handle)

    @mock.patch.object(vhdutils.VHDUtils, '_open')
    def test_reconnect_parent_vhd(self, mock_open):
        set_vdisk_info_struct = (
            self._vdisk_struct.SET_VIRTUAL_DISK_INFO)
        open_params_struct = (
            self._vdisk_struct.OPEN_VIRTUAL_DISK_PARAMETERS)

        fake_set_params = set_vdisk_info_struct.return_value
        fake_open_params = open_params_struct.return_value
        mock_open.return_value = mock.sentinel.handle

        self._vhdutils.reconnect_parent_vhd(mock.sentinel.vhd_path,
                                            mock.sentinel.parent_path)

        self.assertEqual(w_const.OPEN_VIRTUAL_DISK_VERSION_2,
                         fake_open_params.Version)
        self.assertFalse(fake_open_params.Version2.GetInfoOnly)

        self._vhdutils._open.assert_called_once_with(
            mock.sentinel.vhd_path,
            open_flag=w_const.OPEN_VIRTUAL_DISK_FLAG_NO_PARENTS,
            open_access_mask=0,
            open_params=vhdutils.ctypes.byref(fake_open_params))

        self.assertEqual(w_const.SET_VIRTUAL_DISK_INFO_PARENT_PATH,
                         fake_set_params.Version)
        self.assertEqual(mock.sentinel.parent_path,
                         fake_set_params.ParentFilePath)

        self._mock_run.assert_called_once_with(
            vhdutils.virtdisk.SetVirtualDiskInformation,
            mock.sentinel.handle,
            vhdutils.ctypes.byref(fake_set_params),
            **self._run_args)
        self._mock_close.assert_called_once_with(mock.sentinel.handle)

    @mock.patch.object(vhdutils.VHDUtils, 'get_internal_vhd_size_by_file_size')
    @mock.patch.object(vhdutils.VHDUtils, '_resize_vhd')
    @mock.patch.object(vhdutils.VHDUtils, '_check_resize_needed')
    def _test_resize_vhd(self, mock_check_resize_needed,
                         mock_resize_helper, mock_get_internal_size,
                         is_file_max_size=True, resize_needed=True):
        mock_check_resize_needed.return_value = resize_needed

        self._vhdutils.resize_vhd(mock.sentinel.vhd_path,
                                  mock.sentinel.new_size,
                                  is_file_max_size,
                                  validate_new_size=True)

        if is_file_max_size:
            mock_get_internal_size.assert_called_once_with(
                mock.sentinel.vhd_path, mock.sentinel.new_size)
            expected_new_size = mock_get_internal_size.return_value
        else:
            expected_new_size = mock.sentinel.new_size

        mock_check_resize_needed.assert_called_once_with(
            mock.sentinel.vhd_path, expected_new_size)
        if resize_needed:
            mock_resize_helper.assert_called_once_with(mock.sentinel.vhd_path,
                                                       expected_new_size)
        else:
            self.assertFalse(mock_resize_helper.called)

    def test_resize_vhd_specifying_internal_size(self):
        self._test_resize_vhd(is_file_max_size=False)

    def test_resize_vhd_specifying_file_max_size(self):
        self._test_resize_vhd()

    def test_resize_vhd_already_having_requested_size(self):
        self._test_resize_vhd(resize_needed=False)

    @mock.patch.object(vhdutils.VHDUtils, 'get_vhd_size')
    def _test_check_resize_needed(self, mock_get_vhd_size,
                                  current_size=1, new_size=2):
        mock_get_vhd_size.return_value = dict(VirtualSize=current_size)

        if current_size > new_size:
            self.assertRaises(exceptions.VHDException,
                              self._vhdutils._check_resize_needed,
                              mock.sentinel.vhd_path,
                              new_size)
        else:
            resize_needed = self._vhdutils._check_resize_needed(
                mock.sentinel.vhd_path, new_size)
            self.assertEqual(current_size < new_size, resize_needed)

    def test_check_resize_needed_smaller_new_size(self):
        self._test_check_resize_needed(current_size=2, new_size=1)

    def test_check_resize_needed_bigger_new_size(self):
        self._test_check_resize_needed()

    def test_check_resize_needed_smaller_equal_size(self):
        self._test_check_resize_needed(current_size=1, new_size=1)

    @mock.patch.object(vhdutils.VHDUtils, '_open')
    def test_resize_vhd_helper(self, mock_open):
        resize_vdisk_struct = (
            self._vdisk_struct.RESIZE_VIRTUAL_DISK_PARAMETERS)
        fake_params = resize_vdisk_struct.return_value

        mock_open.return_value = mock.sentinel.handle

        self._vhdutils._resize_vhd(mock.sentinel.vhd_path,
                                   mock.sentinel.new_size)

        self.assertEqual(w_const.RESIZE_VIRTUAL_DISK_VERSION_1,
                         fake_params.Version)
        self.assertEqual(mock.sentinel.new_size,
                         fake_params.Version1.NewSize)

        self._mock_run.assert_called_once_with(
            vhdutils.virtdisk.ResizeVirtualDisk,
            mock.sentinel.handle,
            0,
            vhdutils.ctypes.byref(fake_params),
            None,
            **self._run_args)
        self._mock_close.assert_called_once_with(mock.sentinel.handle)

    @mock.patch.object(vhdutils.VHDUtils, 'get_vhd_info')
    @mock.patch.object(vhdutils.VHDUtils,
                       '_get_internal_vhd_size_by_file_size')
    @mock.patch.object(vhdutils.VHDUtils,
                       '_get_internal_vhdx_size_by_file_size')
    def _test_get_int_sz_by_file_size(
            self, mock_get_vhdx_int_size,
            mock_get_vhd_int_size, mock_get_vhd_info,
            vhd_dev_id=w_const.VIRTUAL_STORAGE_TYPE_DEVICE_VHD,
            vhd_type=constants.VHD_TYPE_DYNAMIC):
        fake_vhd_info = dict(ProviderSubtype=vhd_type,
                             ParentPath=mock.sentinel.parent_path,
                             DeviceId=vhd_dev_id)
        mock_get_vhd_info.side_effect = [fake_vhd_info]
        exppected_vhd_info_calls = [mock.call(mock.sentinel.vhd_path)]
        expected_vhd_checked = mock.sentinel.vhd_path
        expected_checked_vhd_info = fake_vhd_info

        if vhd_type == constants.VHD_TYPE_DIFFERENCING:
            expected_checked_vhd_info = dict(
                fake_vhd_info, vhd_type=constants.VHD_TYPE_DYNAMIC)
            mock_get_vhd_info.side_effect.append(
                expected_checked_vhd_info)
            exppected_vhd_info_calls.append(
                mock.call(mock.sentinel.parent_path))
            expected_vhd_checked = mock.sentinel.parent_path

        is_vhd = vhd_dev_id == w_const.VIRTUAL_STORAGE_TYPE_DEVICE_VHD
        expected_helper = (mock_get_vhd_int_size
                           if is_vhd
                           else mock_get_vhdx_int_size)

        ret_val = self._vhdutils.get_internal_vhd_size_by_file_size(
            mock.sentinel.vhd_path, mock.sentinel.vhd_size)

        mock_get_vhd_info.assert_has_calls(exppected_vhd_info_calls)
        expected_helper.assert_called_once_with(expected_vhd_checked,
                                                mock.sentinel.vhd_size,
                                                expected_checked_vhd_info)
        self.assertEqual(expected_helper.return_value, ret_val)

    def test_get_int_sz_by_file_size_vhd(self):
        self._test_get_int_sz_by_file_size()

    def test_get_int_sz_by_file_size_vhdx(self):
        self._test_get_int_sz_by_file_size(
            vhd_dev_id=w_const.VIRTUAL_STORAGE_TYPE_DEVICE_VHDX)

    def test_get_int_sz_by_file_size_differencing(self):
        self._test_get_int_sz_by_file_size(
            vhd_dev_id=w_const.VIRTUAL_STORAGE_TYPE_DEVICE_VHDX)

    def _mocked_get_internal_vhd_size(self, root_vhd_size, vhd_type):
        fake_vhd_info = dict(ProviderSubtype=vhd_type,
                             BlockSize=2097152,
                             ParentPath=mock.sentinel.parent_path)

        return self._vhdutils._get_internal_vhd_size_by_file_size(
            mock.sentinel.vhd_path, root_vhd_size, fake_vhd_info)

    def test_get_internal_vhd_size_by_file_size_fixed(self):
        root_vhd_size = 1 << 30
        real_size = self._mocked_get_internal_vhd_size(
            root_vhd_size=root_vhd_size,
            vhd_type=constants.VHD_TYPE_FIXED)

        expected_vhd_size = root_vhd_size - 512
        self.assertEqual(expected_vhd_size, real_size)

    def test_get_internal_vhd_size_by_file_size_dynamic(self):
        root_vhd_size = 20 << 30
        real_size = self._mocked_get_internal_vhd_size(
            root_vhd_size=root_vhd_size,
            vhd_type=constants.VHD_TYPE_DYNAMIC)

        expected_md_size = 43008
        expected_vhd_size = root_vhd_size - expected_md_size
        self.assertEqual(expected_vhd_size, real_size)

    @mock.patch.object(vhdutils.VHDUtils, '_get_vhdx_block_size')
    @mock.patch.object(vhdutils.VHDUtils, '_get_vhdx_log_size')
    @mock.patch.object(vhdutils.VHDUtils, '_get_vhdx_metadata_size_and_offset')
    def test_get_vhdx_internal_size(self, mock_get_vhdx_md_sz_and_off,
                                    mock_get_vhdx_log_sz,
                                    mock_get_vhdx_block_size):
        self._mock_open()
        fake_log_sz = 1 << 20
        fake_block_sz = 32 << 20
        fake_md_sz = 1 << 20
        fake_logical_sector_sz = 4096
        new_vhd_sz = 1 << 30
        # We expect less than a block to be reserved for internal metadata.
        expected_max_int_sz = new_vhd_sz - fake_block_sz

        fake_vhd_info = dict(SectorSize=fake_logical_sector_sz)

        mock_get_vhdx_block_size.return_value = fake_block_sz
        mock_get_vhdx_log_sz.return_value = fake_log_sz
        mock_get_vhdx_md_sz_and_off.return_value = fake_md_sz, None

        internal_size = self._vhdutils._get_internal_vhdx_size_by_file_size(
            mock.sentinel.vhd_path, new_vhd_sz, fake_vhd_info)

        self.assertIn(type(internal_size), six.integer_types)
        self.assertEqual(expected_max_int_sz, internal_size)

    def test_get_vhdx_internal_size_exception(self):
        mock_open = self._mock_open()
        mock_open.side_effect = IOError
        func = self._vhdutils._get_internal_vhdx_size_by_file_size
        self.assertRaises(exceptions.VHDException,
                          func,
                          mock.sentinel.vhd_path,
                          mock.sentinel.vhd_size,
                          mock.sentinel.vhd_info)

    def _get_mock_file_handle(self, *args):
        mock_file_handle = mock.Mock()
        mock_file_handle.read.side_effect = args
        return mock_file_handle

    def test_get_vhdx_current_header(self):
        # The current header has the maximum sequence number.
        fake_seq_numbers = [
            bytearray(b'\x01\x00\x00\x00\x00\x00\x00\x00'),
            bytearray(b'\x02\x00\x00\x00\x00\x00\x00\x00')]
        mock_handle = self._get_mock_file_handle(*fake_seq_numbers)

        offset = self._vhdutils._get_vhdx_current_header_offset(mock_handle)

        self.assertEqual(vhdutils.VHDX_HEADER_OFFSETS[1], offset)

    @mock.patch.object(vhdutils.VHDUtils, '_get_vhdx_current_header_offset')
    def test_get_log_size(self, mock_get_vhdx_curr_hd_offset):
        fake_curr_header_offset = vhdutils.VHDX_HEADER_OFFSETS[0]
        fake_log_sz = bytearray(b'\x01\x00\x00\x00')

        mock_get_vhdx_curr_hd_offset.return_value = fake_curr_header_offset
        mock_handle = self._get_mock_file_handle(fake_log_sz)

        log_size = self._vhdutils._get_vhdx_log_size(mock_handle)

        self.assertEqual(log_size, 1)

    def test_get_vhdx_metadata_size(self):
        fake_md_offset = bytearray(b'\x01\x00\x00\x00\x00\x00\x00\x00')
        fake_md_sz = bytearray(b'\x01\x00\x00\x00')

        mock_handle = self._get_mock_file_handle(fake_md_offset,
                                                 fake_md_sz)

        md_sz, md_offset = self._vhdutils._get_vhdx_metadata_size_and_offset(
            mock_handle)

        self.assertEqual(1, md_sz)
        self.assertEqual(1, md_offset)

    @mock.patch.object(vhdutils.VHDUtils,
                       '_get_vhdx_metadata_size_and_offset')
    def test_get_block_size(self, mock_get_md_sz_and_offset):
        mock_get_md_sz_and_offset.return_value = (mock.sentinel.md_sz, 1024)
        fake_block_size = bytearray(b'\x01\x00\x00\x00')
        fake_offset = bytearray(b'\x02\x00\x00\x00')
        mock_handle = self._get_mock_file_handle(fake_offset,
                                                 fake_block_size)

        block_size = self._vhdutils._get_vhdx_block_size(mock_handle)
        self.assertEqual(block_size, 1)

    @mock.patch.object(vhdutils.VHDUtils, 'convert_vhd')
    @mock.patch.object(os, 'unlink')
    @mock.patch.object(os, 'rename')
    def test_flatten_vhd(self, mock_rename, mock_unlink, mock_convert):
        fake_vhd_path = r'C:\test.vhd'
        expected_tmp_path = r'C:\test.tmp.vhd'

        self._vhdutils.flatten_vhd(fake_vhd_path)

        mock_convert.assert_called_once_with(fake_vhd_path, expected_tmp_path)
        mock_unlink.assert_called_once_with(fake_vhd_path)
        mock_rename.assert_called_once_with(expected_tmp_path, fake_vhd_path)

    def test_get_best_supported_vhd_format(self):
        fmt = self._vhdutils.get_best_supported_vhd_format()
        self.assertEqual(constants.DISK_FORMAT_VHDX, fmt)

    @ddt.data({},
              {'read_only': False, 'detach_on_handle_close': True})
    @ddt.unpack
    @mock.patch.object(vhdutils.VHDUtils, '_open')
    def test_attach_virtual_disk(self, mock_open, read_only=True,
                                 detach_on_handle_close=False):
        ret_val = self._vhdutils.attach_virtual_disk(
            mock.sentinel.vhd_path,
            read_only, detach_on_handle_close)

        handle = mock_open.return_value
        self.assertEqual(handle
                         if detach_on_handle_close else None,
                         ret_val)

        exp_access_mask = (w_const.VIRTUAL_DISK_ACCESS_ATTACH_RO
                           if read_only
                           else w_const.VIRTUAL_DISK_ACCESS_ATTACH_RW)
        mock_open.assert_called_once_with(mock.sentinel.vhd_path,
                                          open_access_mask=exp_access_mask)

        self._mock_run.assert_called_once_with(
            vhdutils.virtdisk.AttachVirtualDisk,
            handle,
            None,
            mock.ANY,
            0, None, None,
            **self._run_args)

        if not detach_on_handle_close:
            self._mock_close.assert_called_once_with(handle)
        else:
            self._mock_close.assert_not_called()

        mock_run_args = self._mock_run.call_args_list[0][0]
        attach_flag = mock_run_args[3]

        self.assertEqual(
            read_only,
            bool(attach_flag & w_const.ATTACH_VIRTUAL_DISK_FLAG_READ_ONLY))
        self.assertEqual(
            not detach_on_handle_close,
            bool(attach_flag &
                 w_const.ATTACH_VIRTUAL_DISK_FLAG_PERMANENT_LIFETIME))

    @mock.patch.object(vhdutils.VHDUtils, '_open')
    def test_detach_virtual_disk(self, mock_open):
        self._mock_run.return_value = w_const.ERROR_NOT_READY

        self._vhdutils.detach_virtual_disk(mock.sentinel.vhd_path)

        mock_open.assert_called_once_with(
            mock.sentinel.vhd_path,
            open_access_mask=w_const.VIRTUAL_DISK_ACCESS_DETACH)

        self._mock_run.assert_called_once_with(
            vhdutils.virtdisk.DetachVirtualDisk,
            mock_open.return_value,
            0, 0,
            ignored_error_codes=[w_const.ERROR_NOT_READY],
            **self._run_args)
        self._mock_close.assert_called_once_with(mock_open.return_value)

    @mock.patch.object(vhdutils.VHDUtils, '_open')
    def test_get_virtual_disk_physical_path(self, mock_open):
        self._ctypes_patcher.stop()
        vhdutils.wintypes = wintypes

        fake_drive_path = r'\\.\PhysicialDrive5'

        def fake_run(func, handle, disk_path_sz_p, disk_path, **kwargs):
            disk_path_sz = ctypes.cast(
                disk_path_sz_p, wintypes.PULONG).contents.value
            self.assertEqual(w_const.MAX_PATH, disk_path_sz)

            disk_path.value = fake_drive_path

        self._mock_run.side_effect = fake_run

        ret_val = self._vhdutils.get_virtual_disk_physical_path(
            mock.sentinel.vhd_path)

        self.assertEqual(fake_drive_path, ret_val)
        mock_open.assert_called_once_with(
            mock.sentinel.vhd_path,
            open_flag=w_const.OPEN_VIRTUAL_DISK_FLAG_NO_PARENTS,
            open_access_mask=(w_const.VIRTUAL_DISK_ACCESS_GET_INFO |
                              w_const.VIRTUAL_DISK_ACCESS_DETACH))

        self._mock_run.assert_called_once_with(
            vhdutils.virtdisk.GetVirtualDiskPhysicalPath,
            mock_open.return_value,
            mock.ANY,
            mock.ANY,
            **self._run_args)



#  Copyright 2014 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from os_win import constants
from os_win import exceptions
from os_win.tests.unit import test_base
from os_win.utils import hostutils


class FakeCPUSpec(object):
    """Fake CPU Spec for unit tests."""

    Architecture = mock.sentinel.cpu_arch
    Name = mock.sentinel.cpu_name
    Manufacturer = mock.sentinel.cpu_man
    MaxClockSpeed = mock.sentinel.max_clock_speed
    NumberOfCores = mock.sentinel.cpu_cores
    NumberOfLogicalProcessors = mock.sentinel.cpu_procs


class HostUtilsTestCase(test_base.OsWinBaseTestCase):
    """Unit tests for the Hyper-V hostutils class."""

    _DEVICE_ID = "Microsoft:UUID\\0\\0"
    _NODE_ID = "Microsoft:PhysicalNode\\0"

    _FAKE_MEMORY_TOTAL = 1024
    _FAKE_MEMORY_FREE = 512
    _FAKE_DISK_SIZE = 1024
    _FAKE_DISK_FREE = 512
    _FAKE_VERSION_GOOD = '6.2.0'
    _FAKE_VERSION_BAD = '6.1.9'

    def setUp(self):
        self._hostutils = hostutils.HostUtils()
        self._hostutils._conn_cimv2 = mock.MagicMock()
        self._hostutils._conn_scimv2 = mock.MagicMock()
        self._hostutils._conn_attr = mock.MagicMock()
        self._hostutils._netutils_prop = mock.MagicMock()
        self._conn = self._hostutils._conn
        self._conn_scimv2 = self._hostutils._conn_scimv2
        self._netutils = self._hostutils._netutils

        super(HostUtilsTestCase, self).setUp()

    @mock.patch('os_win.utilsfactory.get_networkutils')
    def test_netutils(self, mock_get_networkutils):
        self._hostutils._netutils_prop = None
        self.assertEqual(self._hostutils._netutils,
                         mock_get_networkutils.return_value)

    @mock.patch('os_win.utils.hostutils.kernel32')
    def test_get_host_tick_count64(self, mock_kernel32):
        tick_count64 = "100"
        mock_kernel32.GetTickCount64.return_value = tick_count64
        response = self._hostutils.get_host_tick_count64()
        self.assertEqual(tick_count64, response)

    def test_get_cpus_info(self):
        cpu = mock.MagicMock(spec=FakeCPUSpec)
        self._hostutils._conn_cimv2.query.return_value = [cpu]
        cpu_list = self._hostutils.get_cpus_info()
        self.assertEqual([cpu._mock_children], cpu_list)

    def test_get_memory_info(self):
        memory = mock.MagicMock()
        type(memory).TotalVisibleMemorySize = mock.PropertyMock(
            return_value=self._FAKE_MEMORY_TOTAL)
        type(memory).FreePhysicalMemory = mock.PropertyMock(
            return_value=self._FAKE_MEMORY_FREE)

        self._hostutils._conn_cimv2.query.return_value = [memory]
        total_memory, free_memory = self._hostutils.get_memory_info()

        self.assertEqual(self._FAKE_MEMORY_TOTAL, total_memory)
        self.assertEqual(self._FAKE_MEMORY_FREE, free_memory)

    def test_get_volume_info(self):
        disk = mock.MagicMock()
        type(disk).Size = mock.PropertyMock(return_value=self._FAKE_DISK_SIZE)
        type(disk).FreeSpace = mock.PropertyMock(
            return_value=self._FAKE_DISK_FREE)

        self._hostutils._conn_cimv2.query.return_value = [disk]
        (total_memory, free_memory) = self._hostutils.get_volume_info(
            mock.sentinel.FAKE_DRIVE)

        self.assertEqual(self._FAKE_DISK_SIZE, total_memory)
        self.assertEqual(self._FAKE_DISK_FREE, free_memory)

    def test_check_min_windows_version_true(self):
        self._test_check_min_windows_version(self._FAKE_VERSION_GOOD, True)

    def test_check_min_windows_version_false(self):
        self._test_check_min_windows_version(self._FAKE_VERSION_BAD, False)

    def _test_check_min_windows_version(self, version, expected):
        os = mock.MagicMock()
        os.Version = version
        self._hostutils._conn_cimv2.Win32_OperatingSystem.return_value = [os]
        hostutils.HostUtils._windows_version = None
        self.assertEqual(expected,
                         self._hostutils.check_min_windows_version(6, 2))

    def test_get_windows_version(self):
        os = mock.MagicMock()
        os.Version = self._FAKE_VERSION_GOOD
        self._hostutils._conn_cimv2.Win32_OperatingSystem.return_value = [os]
        hostutils.HostUtils._windows_version = None
        self.assertEqual(self._FAKE_VERSION_GOOD,
                         self._hostutils.get_windows_version())

    @mock.patch('socket.gethostname')
    @mock.patch('os_win._utils.get_ips')
    def test_get_local_ips(self, mock_get_ips, mock_gethostname):
        local_ips = self._hostutils.get_local_ips()

        self.assertEqual(mock_get_ips.return_value, local_ips)
        mock_gethostname.assert_called_once_with()
        mock_get_ips.assert_called_once_with(mock_gethostname.return_value)

    def _test_host_power_action(self, action):
        fake_win32 = mock.MagicMock()
        fake_win32.Win32Shutdown = mock.MagicMock()

        self._hostutils._conn_cimv2.Win32_OperatingSystem.return_value = [
            fake_win32]

        if action == constants.HOST_POWER_ACTION_SHUTDOWN:
            self._hostutils.host_power_action(action)
            fake_win32.Win32Shutdown.assert_called_with(
                self._hostutils._HOST_FORCED_SHUTDOWN)
        elif action == constants.HOST_POWER_ACTION_REBOOT:
            self._hostutils.host_power_action(action)
            fake_win32.Win32Shutdown.assert_called_with(
                self._hostutils._HOST_FORCED_REBOOT)
        else:
            self.assertRaises(NotImplementedError,
                              self._hostutils.host_power_action, action)

    def test_host_shutdown(self):
        self._test_host_power_action(constants.HOST_POWER_ACTION_SHUTDOWN)

    def test_host_reboot(self):
        self._test_host_power_action(constants.HOST_POWER_ACTION_REBOOT)

    def test_host_startup(self):
        self._test_host_power_action(constants.HOST_POWER_ACTION_STARTUP)

    def test_get_supported_vm_types_2012_r2(self):
        with mock.patch.object(self._hostutils,
                               'check_min_windows_version') as mock_check_win:
            mock_check_win.return_value = True
            result = self._hostutils.get_supported_vm_types()
            self.assertEqual([constants.IMAGE_PROP_VM_GEN_1,
                              constants.IMAGE_PROP_VM_GEN_2], result)

    def test_get_supported_vm_types(self):
        with mock.patch.object(self._hostutils,
                               'check_min_windows_version') as mock_check_win:
            mock_check_win.return_value = False
            result = self._hostutils.get_supported_vm_types()
            self.assertEqual([constants.IMAGE_PROP_VM_GEN_1], result)

    def test_check_server_feature(self):
        mock_sv_feature_cls = self._hostutils._conn_cimv2.Win32_ServerFeature
        mock_sv_feature_cls.return_value = [mock.sentinel.sv_feature]

        feature_enabled = self._hostutils.check_server_feature(
            mock.sentinel.feature_id)
        self.assertTrue(feature_enabled)

        mock_sv_feature_cls.assert_called_once_with(
            ID=mock.sentinel.feature_id)

    def test_get_nic_sriov_vfs(self):
        mock_vswitch_sd = mock.Mock()
        mock_hw_offload_sd_bad = mock.Mock(IovVfCapacity=0)
        mock_hw_offload_sd_ok = mock.Mock()
        vswitch_sds_class = self._conn.Msvm_VirtualEthernetSwitchSettingData
        vswitch_sds_class.return_value = [mock_vswitch_sd] * 3
        self._conn.Msvm_EthernetSwitchHardwareOffloadData.side_effect = [
            [mock_hw_offload_sd_bad], [mock_hw_offload_sd_ok],
            [mock_hw_offload_sd_ok]]
        self._netutils.get_vswitch_external_network_name.side_effect = [
            None, mock.sentinel.nic_name]
        mock_nic = mock.Mock()
        self._conn_scimv2.MSFT_NetAdapter.return_value = [mock_nic]

        vfs = self._hostutils.get_nic_sriov_vfs()

        expected = {
            'vswitch_name': mock_vswitch_sd.ElementName,
            'device_id': mock_nic.PnPDeviceID,
            'total_vfs': mock_hw_offload_sd_ok.IovVfCapacity,
            'used_vfs': mock_hw_offload_sd_ok.IovVfUsage,
        }
        self.assertEqual([expected], vfs)
        vswitch_sds_class.assert_called_once_with(IOVPreferred=True)
        self._conn.Msvm_EthernetSwitchHardwareOffloadData.assert_has_calls([
            mock.call(SystemName=mock_vswitch_sd.VirtualSystemIdentifier)] * 3)
        self._netutils.get_vswitch_external_network_name.assert_has_calls([
            mock.call(mock_vswitch_sd.ElementName)] * 2)
        self._conn_scimv2.MSFT_NetAdapter.assert_called_once_with(
            InterfaceDescription=mock.sentinel.nic_name)

    @mock.patch.object(hostutils.HostUtils, '_get_nic_hw_offload_info')
    def test_get_nic_hardware_offload_info(self, mock_get_nic_offload):
        mock_vswitch_sd = mock.Mock(VirtualSystemIdentifier=mock.sentinel.vsid)
        mock_hw_offload_sd = mock.Mock(SystemName=mock.sentinel.vsid)

        vswitch_sds_class = self._conn.Msvm_VirtualEthernetSwitchSettingData
        vswitch_sds_class.return_value = [mock_vswitch_sd]
        hw_offload_class = self._conn.Msvm_EthernetSwitchHardwareOffloadData
        hw_offload_class.return_value = [mock_hw_offload_sd]

        hw_offload_info = self._hostutils.get_nic_hardware_offload_info()

        self.assertEqual([mock_get_nic_offload.return_value], hw_offload_info)
        vswitch_sds_class.assert_called_once_with()
        hw_offload_class.assert_called_once_with()
        mock_get_nic_offload.assert_called_once_with(mock_vswitch_sd,
                                                     mock_hw_offload_sd)

    def test_get_nic_hardware_offload_info_no_nic(self):
        self._netutils.get_vswitch_external_network_name.return_value = None
        mock_vswitch_sd = mock.Mock()

        hw_offload_info = self._hostutils._get_nic_hw_offload_info(
            mock_vswitch_sd, mock.sentinel.hw_offload_sd)

        self.assertIsNone(hw_offload_info)

    @mock.patch.object(hostutils.LOG, 'warning')
    def test_get_nic_hw_offload_info(self, mock_warning):
        mock_vswitch_sd = mock.Mock()
        mock_hw_offload_sd = mock.Mock(IovVfCapacity=0)
        mock_nic = mock.Mock()
        self._conn_scimv2.MSFT_NetAdapter.return_value = [mock_nic]

        hw_offload_info = self._hostutils._get_nic_hw_offload_info(
            mock_vswitch_sd, mock_hw_offload_sd)

        expected = {
            'vswitch_name': mock_vswitch_sd.ElementName,
            'device_id': mock_nic.PnPDeviceID,
            'total_vfs': mock_hw_offload_sd.IovVfCapacity,
            'used_vfs': mock_hw_offload_sd.IovVfUsage,
            'total_iov_queue_pairs': mock_hw_offload_sd.IovQueuePairCapacity,
            'used_iov_queue_pairs': mock_hw_offload_sd.IovQueuePairUsage,
            'total_vmqs': mock_hw_offload_sd.VmqCapacity,
            'used_vmqs': mock_hw_offload_sd.VmqUsage,
            'total_ipsecsa': mock_hw_offload_sd.IPsecSACapacity,
            'used_ipsecsa': mock_hw_offload_sd.IPsecSAUsage,
        }
        self.assertEqual(expected, hw_offload_info)
        get_ext_net_name = self._netutils.get_vswitch_external_network_name
        get_ext_net_name.assert_called_once_with(mock_vswitch_sd.ElementName)
        self.assertTrue(mock_warning.called)
        self._conn_scimv2.MSFT_NetAdapter.assert_called_once_with(
            InterfaceDescription=get_ext_net_name.return_value)

    def _check_get_numa_nodes_missing_info(self):
        numa_node = mock.MagicMock()
        self._hostutils._conn.Msvm_NumaNode.return_value = [
            numa_node, numa_node]

        nodes_info = self._hostutils.get_numa_nodes()
        self.assertEqual([], nodes_info)

    @mock.patch.object(hostutils.HostUtils, '_get_numa_memory_info')
    def test_get_numa_nodes_missing_memory_info(self, mock_get_memory_info):
        mock_get_memory_info.return_value = None
        self._check_get_numa_nodes_missing_info()

    @mock.patch.object(hostutils.HostUtils, '_get_numa_cpu_info')
    @mock.patch.object(hostutils.HostUtils, '_get_numa_memory_info')
    def test_get_numa_nodes_missing_cpu_info(self, mock_get_memory_info,
                                             mock_get_cpu_info):
        mock_get_cpu_info.return_value = None
        self._check_get_numa_nodes_missing_info()

    @mock.patch.object(hostutils.HostUtils, '_get_numa_cpu_info')
    @mock.patch.object(hostutils.HostUtils, '_get_numa_memory_info')
    def test_get_numa_nodes(self, mock_get_memory_info, mock_get_cpu_info):
        numa_memory = mock_get_memory_info.return_value
        host_cpu = mock.MagicMock(DeviceID=self._DEVICE_ID)
        mock_get_cpu_info.return_value = [host_cpu]
        numa_node = mock.MagicMock(NodeID=self._NODE_ID)
        self._hostutils._conn.Msvm_NumaNode.return_value = [
            numa_node, numa_node]

        nodes_info = self._hostutils.get_numa_nodes()

        expected_info = {
            'id': self._DEVICE_ID.split('\\')[-1],
            'memory': numa_memory.NumberOfBlocks,
            'memory_usage': numa_node.CurrentlyConsumableMemoryBlocks,
            'cpuset': set([self._DEVICE_ID.split('\\')[-1]]),
            'cpu_usage': 0,
        }

        self.assertEqual([expected_info, expected_info], nodes_info)

    def test_get_numa_memory_info(self):
        system_memory = mock.MagicMock()
        system_memory.path_.return_value = 'fake_wmi_obj_path'
        numa_node_memory = mock.MagicMock()
        numa_node_memory.path_.return_value = 'fake_wmi_obj_path1'
        numa_node_assoc = [system_memory]
        memory_info = self._hostutils._get_numa_memory_info(
            numa_node_assoc, [system_memory, numa_node_memory])

        self.assertEqual(system_memory, memory_info)

    def test_get_numa_memory_info_not_found(self):
        other = mock.MagicMock()
        memory_info = self._hostutils._get_numa_memory_info([], [other])

        self.assertIsNone(memory_info)

    def test_get_numa_cpu_info(self):
        host_cpu = mock.MagicMock()
        host_cpu.path_.return_value = 'fake_wmi_obj_path'
        vm_cpu = mock.MagicMock()
        vm_cpu.path_.return_value = 'fake_wmi_obj_path1'
        numa_node_assoc = [host_cpu]
        cpu_info = self._hostutils._get_numa_cpu_info(numa_node_assoc,
                                                      [host_cpu, vm_cpu])

        self.assertEqual([host_cpu], cpu_info)

    def test_get_numa_cpu_info_not_found(self):
        other = mock.MagicMock()
        cpu_info = self._hostutils._get_numa_cpu_info([], [other])

        self.assertEqual([], cpu_info)

    def test_get_remotefx_gpu_info(self):
        fake_gpu = mock.MagicMock()
        fake_gpu.Name = mock.sentinel.Fake_gpu_name
        fake_gpu.TotalVideoMemory = mock.sentinel.Fake_gpu_total_memory
        fake_gpu.AvailableVideoMemory = mock.sentinel.Fake_gpu_available_memory
        fake_gpu.DirectXVersion = mock.sentinel.Fake_gpu_directx
        fake_gpu.DriverVersion = mock.sentinel.Fake_gpu_driver_version

        mock_phys_3d_proc = (
            self._hostutils._conn.Msvm_Physical3dGraphicsProcessor)
        mock_phys_3d_proc.return_value = [fake_gpu]

        return_gpus = self._hostutils.get_remotefx_gpu_info()
        self.assertEqual(mock.sentinel.Fake_gpu_name, return_gpus[0]['name'])
        self.assertEqual(mock.sentinel.Fake_gpu_driver_version,
                         return_gpus[0]['driver_version'])
        self.assertEqual(mock.sentinel.Fake_gpu_total_memory,
                         return_gpus[0]['total_video_ram'])
        self.assertEqual(mock.sentinel.Fake_gpu_available_memory,
                         return_gpus[0]['available_video_ram'])
        self.assertEqual(mock.sentinel.Fake_gpu_directx,
                         return_gpus[0]['directx_version'])

    def _set_verify_host_remotefx_capability_mocks(self, isGpuCapable=True,
                                                   isSlatCapable=True):
        s3d_video_pool = self._hostutils._conn.Msvm_Synth3dVideoPool()[0]
        s3d_video_pool.IsGpuCapable = isGpuCapable
        s3d_video_pool.IsSlatCapable = isSlatCapable

    def test_verify_host_remotefx_capability_unsupported_gpu(self):
        self._set_verify_host_remotefx_capability_mocks(isGpuCapable=False)
        self.assertRaises(exceptions.HyperVRemoteFXException,
                          self._hostutils.verify_host_remotefx_capability)

    def test_verify_host_remotefx_capability_no_slat(self):
        self._set_verify_host_remotefx_capability_mocks(isSlatCapable=False)
        self.assertRaises(exceptions.HyperVRemoteFXException,
                          self._hostutils.verify_host_remotefx_capability)

    def test_verify_host_remotefx_capability(self):
        self._set_verify_host_remotefx_capability_mocks()
        self._hostutils.verify_host_remotefx_capability()

    def test_supports_nested_virtualization(self):
        self.assertFalse(self._hostutils.supports_nested_virtualization())

    def test_get_pci_passthrough_devices(self):
        self.assertEqual([], self._hostutils.get_pci_passthrough_devices())



# Copyright 2016 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from os_win._i18n import _
from os_win import exceptions


def get_element_associated_class(conn, class_name, element_instance_id=None,
                                 element_uuid=None, fields=None):
    """Returns the objects associated to an element as a list.

    :param conn: connection to be used to execute the query
    :param class_name: object's class type name to be retrieved
    :param element_instance_id: element class InstanceID
    :param element_uuid: UUID of the element
    :param fields: specific class attributes to be retrieved
    """
    if element_instance_id:
        instance_id = element_instance_id
    elif element_uuid:
        instance_id = "Microsoft:%s" % element_uuid
    else:
        err_msg = _("Could not get element associated class. Either element "
                    "instance id or element uuid must be specified.")
        raise exceptions.WqlException(err_msg)
    fields = ", ".join(fields) if fields else "*"
    return conn.query(
        "SELECT %(fields)s FROM %(class_name)s WHERE InstanceID "
        "LIKE '%(instance_id)s%%'" % {
            'fields': fields,
            'class_name': class_name,
            'instance_id': instance_id})



# Copyright 2017 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ctypes

from os_win.utils.winapi import wintypes

lib_handle = None


class TRUSTEE(ctypes.Structure):
    _fields_ = [
        ('pMultipleTrustee', wintypes.PVOID),
        ('MultipleTrusteeOperation', wintypes.INT),
        ('TrusteeForm', wintypes.INT),
        ('TrusteeType', wintypes.INT),
        ('pstrName', wintypes.LPWSTR)
    ]


class EXPLICIT_ACCESS(ctypes.Structure):
    _fields_ = [
        ('grfAccessPermissions', wintypes.DWORD),
        ('grfAccessMode', wintypes.INT),
        ('grfInheritance', wintypes.DWORD),
        ('Trustee', TRUSTEE)
    ]


PEXPLICIT_ACCESS = ctypes.POINTER(EXPLICIT_ACCESS)


def register():
    global lib_handle
    lib_handle = ctypes.windll.advapi32

    lib_handle.GetNamedSecurityInfoW.argtypes = [
        wintypes.LPCWSTR,
        wintypes.INT,
        wintypes.DWORD,
        wintypes.PVOID,
        wintypes.PVOID,
        wintypes.PVOID,
        wintypes.PVOID,
        wintypes.PVOID
    ]
    lib_handle.GetNamedSecurityInfoW.restype = wintypes.DWORD

    lib_handle.SetEntriesInAclW.argtypes = [
        wintypes.ULONG,
        PEXPLICIT_ACCESS,
        wintypes.PVOID,
        wintypes.PVOID
    ]
    lib_handle.SetEntriesInAclW.restype = wintypes.DWORD

    lib_handle.SetNamedSecurityInfoW.argtypes = [
        wintypes.LPWSTR,
        wintypes.INT,
        wintypes.DWORD,
        wintypes.PVOID,
        wintypes.PVOID,
        wintypes.PVOID,
        wintypes.PVOID
    ]
    lib_handle.SetNamedSecurityInfoW.restype = wintypes.DWORD
